# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ()

import argparse
import dataclasses
import sys
import textwrap
from collections.abc import Sequence
from contextlib import ExitStack

from lsst.daf.butler import Butler
from lsst.resources import ResourcePath

from ..pipeline import Pipeline
from ._pipeline_graph import PipelineGraph, TaskImportMode
from .visualization._options import NodeAttributeOptions
from .visualization._show import show


def main(argv: Sequence[str]) -> int:
    """Run the development command-line interface.

    Parameters
    ----------
    argv : `~collections.abc.Sequence` [ `str` ]
        Commmand-line arguments, not including the program (typically
        ``sys.argv[1:]``).

    Notes
    -----
    This CLI is much more capable than pipetask's for very specific things, but
    not as polished or user friendly.  It is intended primarily for use during
    development of pipeline graph (especially visualization, where tests can't
    really be used to judge the quality of the outputs, and a fast turnaround
    is desirable).
    """
    parser = argparse.ArgumentParser(
        description="Expand, resolve, and display pipelines as graphs.",
        epilog=textwrap.dedent(
            """
            WARNING: This is an experimental/development command-line interface
            that is subject to change or removal without warning. The `--show`
            option to 'pipetask build' (with 'pipeline-graph', 'task-graph', or
            'dataset-type-graph' as the argument) is the preferred way to
            display text-based pipeline graphs.
            """
        ),
    )
    Arguments.add_args_to_parser(parser)
    args = Arguments.from_parsed_args(parser.parse_args(argv))
    pipeline_graph = read_input_pipeline(args.input_pipeline)
    if args.resolve:
        butler = Butler.from_config(args.resolve, writeable=False)
        pipeline_graph.resolve(butler.registry)
    if args.save:
        pipeline_graph._write_uri(ResourcePath(args.save))
    if args.show:
        with ExitStack() as stack:
            if args.show == "-":
                stream = sys.stdout
            else:
                path = ResourcePath(args.show)
                # ResourcePath.open doesn't have the kind of mode overloading
                # necessary to make the return type clear enough for MyPy.
                stream = stack.enter_context(path.open("r"))  # type: ignore
            show(
                pipeline_graph,
                stream,
                dataset_types=args.display.dataset_types,
                init=args.display.init,
                color=args.display.color,
                dimensions=args.display.node_attributes.dimensions,
                task_classes=args.display.node_attributes.task_classes,  # type: ignore
                storage_classes=args.display.node_attributes.storage_classes,
                merge_input_trees=args.display.merge_input_trees,
                merge_output_trees=args.display.merge_output_trees,
                merge_intermediates=args.display.merge_intermediates,
                include_automatic_connections=args.display.include_automatic_connections,
                width=args.display.width,
                column_crossing_penalty=args.display.column_crossing_penalty,
                column_insertion_penalty=args.display.column_insertion_penalty,
                column_interior_penalty=args.display.column_interior_penalty,
            )
    return 0


def read_input_pipeline(uri: str) -> PipelineGraph:
    """Read an input pipeline or pipeline graph from a URI.

    Parameters
    ----------
    uri : `str`
        URI to read.  Extension is used to determine whether this is a pipeline
        (.yaml) or pipeline graph (.json.gz).

    Returns
    -------
    graph : `.PipelineGraph`
        Pipeline graph.
    """
    path = ResourcePath(uri)
    match path.getExtension():
        case ".yaml":
            pipeline = Pipeline.from_uri(path)
            return pipeline.to_graph()
        case ".json.gz":
            return PipelineGraph._read_uri(path, import_mode=TaskImportMode.DO_NOT_IMPORT)
        case other:
            raise ValueError(f"Unexpected extension for pipeline file: {other!r}.")


@dataclasses.dataclass
class Arguments:
    """Struct that manages the CLI arguments and options."""

    input_pipeline: str
    """URI to the input pipeline."""

    save: str | None
    """URI for the saved pipeline, or `None` if it will not be saved."""

    show: str | None
    """File to write the graph visualization to; ``-`` for STDOUT, or `None`
    for no visualization.
    """

    resolve: str | None
    """Butler repository URI to use to resolve the graph."""

    display: DisplayArguments
    """Additional options specific to visualization."""

    @classmethod
    def from_parsed_args(cls, args: argparse.Namespace) -> Arguments:
        """Interpret parsed arguments.

        Parameters
        ----------
        args : `argparse.Namespace`
            Parsed argument struct, as returned by
            `argparse.ArgumentParser.parse_args`.

        Returns
        -------
        arguments : `Arguments`
            Intepreted arguments struct.
        """
        return cls(
            input_pipeline=args.input_pipeline,
            save=args.save,
            show=args.show,
            resolve=args.resolve,
            display=DisplayArguments.from_parsed_args(args),
        )

    @classmethod
    def add_args_to_parser(cls, parser: argparse.ArgumentParser) -> None:
        """Add the options and arguments described by this class to a parser.

        Parameters
        ----------
        parser : `argparse.ArgumentParser`
            Argument parser to modify in-place.
        """
        parser.add_argument(
            "input_pipeline",
            type=str,
            metavar="URI",
            help="""
                Filename or URI for the input pipeline specification (.yaml) or
                graph (.json.gz) file to read.
            """,
        )
        parser.add_argument(
            "--save",
            type=str,
            metavar="FILE",
            help="""
                Save the pipeline graph content to this file.  Should have a
                .json.gz extension or no extension (in which case .json.gz will
                be added).
            """,
            default=None,
        )
        parser.add_argument(
            "--show",
            type=str,
            nargs="?",
            metavar="FILE",
            help="""
                Print the pipeline graph in human-readable form using unicode
                lines and symbols.  May be '-' or have no value for STDOUT.
            """,
            default=None,
            const="-",
        )
        parser.add_argument(
            "--resolve",
            type=str,
            nargs=1,
            metavar="REPO",
            help="""
                A butler data repository to use to resolve the graph's dataset
                types and dimensions.
            """,
            default=None,
        )
        DisplayArguments.add_args_to_parser(parser)


@dataclasses.dataclass
class DisplayArguments:
    """Struct that manages the CLI arguments and options specific to
    visualization.
    """

    dataset_types: bool
    """Whether the visualization should show dataset type nodes."""

    init: bool | None
    """Whether to include task initialization its inputs and outputs.

    `False` shows only runtime nodes.  `True` shows all init nodes only.
    `None` shows all nodes.
    """

    color: Sequence[str] | bool | None
    """Colors to use for node symbols and edge lines.

    See the `show` argument of the same name for details.
    """

    node_attributes: NodeAttributeOptions
    """Options for which attributes of ndoes to display and simplify."""

    merge_input_trees: int
    """Whether/how to merge input trees with the same structure.

    See the `show` argument of the same name for details.
    """

    merge_output_trees: int
    """Whether/how to merge output trees with the same structure.

    See the `show` argument of the same name for details.
    """

    merge_intermediates: bool
    """Whether to merge internal parallel subgraphs.

    See the `show` argument of the same name for details.
    """

    include_automatic_connections: bool
    """Whether to include automatic connections like config, metadata, and
    logs.
    """

    width: int
    """Width of the graph in columns.

    See the `show` argument of the same name for details.
    """

    column_crossing_penalty: int
    """Graph layout tuning parameter; see the `show` argument of the same name
    for details.
    """

    column_insertion_penalty: int
    """Graph layout tuning parameter; see the `show` argument of the same name
    for details.
    """

    column_interior_penalty: int
    """Graph layout tuning parameter; see the `show` argument of the same name
    for details.
    """

    def __post_init__(self) -> None:
        if self.node_attributes.storage_classes and not self.dataset_types:
            raise argparse.ArgumentError(
                None,
                "--storage-classes does nothing unless --dataset-types or --only-dataset-types is passed.",
            )

    @classmethod
    def from_parsed_args(cls, args: argparse.Namespace) -> DisplayArguments:
        """Interpret parsed arguments.

        Parameters
        ----------
        args : `argparse.Namespace`
            Parsed argument struct, as returned by
            `argparse.ArgumentParser.parse_args`.

        Returns
        -------
        arguments : `Arguments`
            Intepreted arguments struct.
        """
        return cls(
            dataset_types=args.dataset_types or args.dataset_types_only,
            init=args.init,
            color=args.color,
            node_attributes=NodeAttributeOptions(
                dimensions=args.dimensions,
                task_classes=args.task_classes,
                storage_classes=args.storage_classes,
            ),
            merge_input_trees=args.merge_input_trees,
            merge_output_trees=args.merge_output_trees,
            merge_intermediates=args.merge_intermediates,
            include_automatic_connections=args.include_automatic_connections,
            width=args.width,
            column_crossing_penalty=args.column_crossing_penalty,
            column_insertion_penalty=args.column_insertion_penalty,
            column_interior_penalty=args.column_interior_penalty,
        )

    @classmethod
    def add_args_to_parser(cls, parser: argparse.ArgumentParser) -> None:
        """Add the options and arguments described by this class to a parser.

        Parameters
        ----------
        parser : `argparse.ArgumentParser`
            Argument parser to modify in-place.
        """
        group = parser.add_argument_group("additional options for --show")
        dataset_type_inclusion = group.add_mutually_exclusive_group()
        dataset_type_inclusion.add_argument(
            "--dataset-types",
            action="store_true",
            help="Show a graph containing both dataset types and tasks. Default is a task-only graph.",
        )
        group.add_argument(
            "--init-only",
            action="store_true",
            dest="init",
            help="""
                Show a graph of init-input and init-output dataset types and/or
                task initializations instead of the usual runtime graph.
            """,
            default=False,
        )
        group.add_argument(
            "--init",
            action="store_const",
            const=None,
            help="""
                Show a graph of init-input and init-output dataset types and
                task initializations in addition to the usual runtime graph.
                Requires --dataset-types.
            """,
            default=False,
        )
        color_group = group.add_mutually_exclusive_group()
        color_group.add_argument(
            "--color",
            action="store_true",
            help="""
                Always use terminal escape codes to add color to the graph.
                Default is to use color only if an interactive terminal is
                detected.
            """,
            default=None,
        )
        color_group.add_argument(
            "--no-color",
            action="store_false",
            help="""
                Never use terminal escape codes to add color to the graph.
                Default is to use color only if an interactive terminal is
                detected.
            """,
            default=None,
        )
        color_group.add_argument(
            "--palette",
            type=str,
            nargs="+",
            metavar="COLORS",
            help="""
                A list of colors to use for nodes.  Options include 'red',
                'green', 'blue', 'cyan', 'yellow', 'magenta', and any of these
                preceded by 'light' (case insensitive).  Implies --color.
            """,
            dest="color",
        )
        dimensions_group = group.add_mutually_exclusive_group()
        dimensions_group.add_argument(
            "--no-dimensions",
            action="store_false",
            help="""
                Do not include dimensions in node descriptions or merge
                comparisons at all.  This is the default if the loaded graph
                was not resolved and --resolve was not passed.
            """,
            dest="dimensions",
            default=None,
        )
        dimensions_group.add_argument(
            "--full-dimensions",
            action="store_const",
            help="""
                Show full dimensions in node descriptions, including those that
                are implied or required by another dimension in the set.
            """,
            dest="dimensions",
            const="full",
        )
        dimensions_group.add_argument(
            "--concise-dimensions",
            action="store_const",
            help="""
                Show concise dimensions in node descriptions, removing those
                that are implied or required by another dimension in the set.
                This is the default if the loaded graph was already resolved
                or --resolve is passed.
            """,
            dest="dimensions",
            const="concise",
        )
        task_classes_group = group.add_mutually_exclusive_group()
        dimensions_group.add_argument(
            "--no-task-classes",
            action="store_false",
            help="""
                Do not include task classes in node descriptions or merge
                comparisons at all.
            """,
            dest="task_classes",
            default=None,
        )
        task_classes_group.add_argument(
            "--full-task-classes",
            action="store_const",
            help="""
                Show fully-qualified task classes in task node descriptions,
                and use task classes in merge comparisons.
            """,
            dest="task_classes",
            const="full",
        )
        task_classes_group.add_argument(
            "--concise-task-classes",
            action="store_const",
            help="""
                Show unqualified task classes in task node descriptions, and
                use task classes in merge comparisons.  This is the default.
            """,
            const="concise",
            dest="task_classes",
        )
        storage_classes_group = group.add_mutually_exclusive_group()
        storage_classes_group.add_argument(
            "--no-storage-classes",
            action="store_false",
            help="""
                Show storage classes in dataset type node descriptions, and use
                storage classes in merge comparisons.  This is the default if
                the loaded graph was not resolved and --resolve was not passed.
            """,
            dest="storage_classes",
            default=None,
        )
        storage_classes_group.add_argument(
            "--storage-classes",
            action="store_true",
            help="""
                Show storage classes in dataset type node descriptions, and use
                storage classes in merge comparisons.  This is the default if
                the loaded graph was already resolved or --resolve is passed.
            """,
            dest="storage_classes",
            default=None,
        )
        group.add_argument(
            "--merge-input-trees",
            type=int,
            default=4,
            help="""
                Depth at which to merge input trees with the same outputs,
                dimensions, task classes, and storage classes.  Zero disables
                merging.
            """,
        )
        group.add_argument(
            "--merge-output-trees",
            type=int,
            default=4,
            help="""
                Depth at which to merge output trees with the same inputs,
                dimensions, task classes, and storage classes.  Zero disables
                merging.
            """,
        )
        group.add_argument(
            "--no-merge-intermediates",
            action="store_false",
            dest="merge_intermediates",
            help="""
                Disable merging of intermediate nodes that share the same
                inputs, outputs, dimensions, task classes, and storage classes.
            """,
        )
        group.add_argument(
            "--include-automatic-connections",
            action="store_true",
            help="""
                Include output datasets added by the execution system, such
                as configs, metadata, and logs.
            """,
        )
        group.add_argument(
            "--width",
            type=int,
            default=-1,
            help="""
                Width in characters for the graph and node descriptions.
                Default (-1) is to use the terminal width.  May be 0 to put no
                limit on the width.  This only sets whether node descriptions
                are truncated and moved below the graph, so it may be exceeded
                by the graph itself.
            """,
        )
        group.add_argument(
            "--column-crossing-penalty",
            type=int,
            default=1,
            help="""
                When selecting the column for a new node, penalize a
                candidate column by multiplying the number of ongoing vertical
                edges this node's horizontal incoming edges would have to 'hop'
                by this value.
            """,
        )
        group.add_argument(
            "--column-insertion-penalty",
            type=int,
            default=2,
            help="""
                When selecting the column for a new node, penalize adding new
                columns by this amount.
            """,
        )
        group.add_argument(
            "--column-interior-penalty",
            type=int,
            default=1,
            help="""
                When selecting the column for a new node, penalize adding new
                columns between two existing columns by this amount (in
                addition to the --column-insertion-penalty applied to all new
                columns).
            """,
        )


if __name__ == "__main__":
    main(sys.argv[1:])



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Symbols defined in this package should be imported from
`all_dimensions_quantum_graph_builder` instead; it only appears in the docs
due to limitations in Sphinx.
"""

from __future__ import annotations

__all__ = ("DatasetQueryConstraintVariant",)

import sys
import warnings
from collections.abc import Iterable, Iterator
from typing import Protocol

from lsst.utils.introspection import find_outside_stacklevel


class DatasetQueryConstraintVariant(Iterable, Protocol):
    """Base for all the valid variants for controlling
    constraining graph building queries based on dataset type existence.

    ALL variant corresponds to using all input dataset types to constrain
    a query.

    OFF variant corresponds to not using any dataset types to constrain a
    graph building query.

    LIST variant should be used when one or more specific names should be used
    in constraining a graph building query.

    Normally the ALL and OFF variants are used as as Singletons, attempting to
    instantiate them (i.e. ALL()) will return in singleton class itself.

    LIST is used as a constructor to the contents (i.e. List(['a', 'b'])).
    Using the LIST variant directly as a singleton will behave the same as if
    it were an empty instance.

    Variants can be directly used, or automatically be selected by using the
    `fromExpression` class method given a valid string.
    """

    ALL: "type[_ALL]"
    OFF: "type[_OFF]"
    LIST: "type[_LIST]"

    @classmethod
    def __subclasshook__(cls, subclass):
        if subclass == cls.ALL or subclass == cls.OFF or subclass == cls.LIST:
            return True
        return False

    @classmethod
    def fromExpression(cls, expression: str) -> "DatasetQueryConstraintVariant":
        """Select and return the correct Variant that corresponds to the input
        expression.

        Parameters
        ----------
        expression : `str`
            Input expression. Valid values are ``all`` for all inputs dataset
            types in pipeline, ``off`` to not consider dataset type existence
            as a constraint, single or comma-separated list of dataset type
            names.

        Returns
        -------
        variant : `DatasetQueryConstraintVariant`
            Correct variant for this expression.
        """
        if not isinstance(expression, str):
            raise ValueError("Expression must be a string")
        elif expression == "all":
            return cls.ALL
        elif expression == "off":
            return cls.OFF
        else:
            if " " in expression:
                warnings.warn(
                    "Whitespace found in expression will be trimmed",
                    RuntimeWarning,
                    stacklevel=find_outside_stacklevel("lsst.pipe.base"),
                )
                expression = expression.replace(" ", "")
            members = expression.split(",")
            return cls.LIST(members)


if sys.version_info.minor < 12:
    MetaMeta = type
else:

    class MetaMeta(type(Protocol)):
        def __init__(cls, *args, **kwargs):
            return super().__init__(cls, *args, **kwargs)


class _ALLMETA(DatasetQueryConstraintVariant, type(Protocol), metaclass=MetaMeta):
    def __iter__(self) -> Iterator:  # noqa: N804
        raise NotImplementedError("This variant cannot be iterated")


class _ALL(metaclass=_ALLMETA):
    def __new__(cls):
        return cls


class _OFFMETA(DatasetQueryConstraintVariant, type(Protocol), metaclass=MetaMeta):
    def __iter__(self) -> Iterator:  # noqa: N804
        raise NotImplementedError("This variant cannot be iterated")


class _OFF(metaclass=_OFFMETA):
    def __new__(cls):
        return cls


class _LISTMETA(type(Protocol), metaclass=MetaMeta):
    def __iter__(self):  # noqa: N804
        return iter(tuple())

    def __len__(self):  # noqa: N804
        return 0

    def __eq__(self, o: object) -> bool:  # noqa: N804
        if isinstance(o, self):
            return True
        return super().__eq__(o)


class _LIST(DatasetQueryConstraintVariant, metaclass=_LISTMETA):
    def __init__(self, members: Iterable[str]):
        self.members = list(members)

    def __len__(self) -> int:
        return len(self.members)

    def __iter__(self) -> Iterable[str]:
        return iter(self.members)

    def __repr__(self) -> str:
        return repr(self.members)

    def __str__(self) -> str:
        return str(self.members)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, type(self)):
            return self.members == o.members
        return super().__eq__(o)


def suppressInit(self):
    raise NotImplementedError(
        "DatasetQueryConstraintVariants cannot be directly instantiated. "
        "Please use the variants or the fromExpression class method"
    )


DatasetQueryConstraintVariant.__init__ = suppressInit
DatasetQueryConstraintVariant.ALL = _ALL
DatasetQueryConstraintVariant.OFF = _OFF
DatasetQueryConstraintVariant.LIST = _LIST



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Tests for execution butler."""

import os
import unittest

import lsst.utils.tests
from lsst.daf.butler import Butler
from lsst.daf.butler.direct_butler import DirectButler
from lsst.pipe.base.executionButlerBuilder import buildExecutionButler
from lsst.pipe.base.tests import simpleQGraph
from lsst.utils.tests import temporaryDirectory

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class ExecutionTestCase(unittest.TestCase):
    """Small bunch of tests for instantiating execution butler."""

    def test_simple(self) -> None:
        """Test buildExecutionButler with default butler config."""
        with temporaryDirectory() as root:
            butler, qgraph = simpleQGraph.makeSimpleQGraph(root=root, inMemory=False)
            assert isinstance(butler, DirectButler), "Expect DirectButler in tests"
            buildExecutionButler(butler, qgraph, os.path.join(root, "exec_butler"), butler.run)

    def test_obscore(self) -> None:
        """Test buildExecutionButler with obscore table in the butler."""
        with temporaryDirectory() as root:
            config = os.path.join(TESTDIR, "config/butler-obscore.yaml")
            butler = simpleQGraph.makeSimpleButler(root=root, inMemory=False, config=config)
            butler, qgraph = simpleQGraph.makeSimpleQGraph(butler=butler)

            # Need a fresh butler instance to make sure it reads config from
            # butler.yaml, which does not have full obscore config.
            butler = Butler.from_config(root, run=butler.run)
            assert isinstance(butler, DirectButler), "Expect DirectButler in tests"
            buildExecutionButler(butler, qgraph, os.path.join(root, "exec_butler"), butler.run)


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining connection types to be used within a
`PipelineTaskConnections` class.
"""

__all__ = ["InitInput", "InitOutput", "Input", "PrerequisiteInput", "Output", "BaseConnection"]

import dataclasses
from collections.abc import Callable, Iterable, Sequence
from typing import ClassVar

from deprecated.sphinx import deprecated as deprecated_sphinx  # avoid clash with BaseConnection.deprecated
from lsst.daf.butler import DataCoordinate, DatasetRef, DatasetType, DimensionUniverse, Registry, StorageClass
from lsst.utils.introspection import find_outside_stacklevel


@dataclasses.dataclass(frozen=True)
class BaseConnection:
    """Base class used for declaring `PipelineTask` connections.

    Attributes
    ----------
    name : `str`
        The name used to identify the dataset type.
    storageClass : `str`
        The storage class used when (un)/persisting the dataset type.
    multiple : `bool`
        Indicates if this connection should expect to contain multiple objects
        of the given dataset type.  Tasks with more than one connection with
        ``multiple=True`` with the same dimensions may want to implement
        `PipelineTaskConnections.adjustQuantum` to ensure those datasets are
        consistent (i.e. zip-iterable) in `PipelineTask.runQuantum()` and
        notify the execution system as early as possible of outputs that will
        not be produced because the corresponding input is missing.
    deprecated : `str`, optional
        A description of why this connection is deprecated, including the
        version after which it may be removed.

        If not `None`, the string is appended to the docstring for this
        connection and the corresponding config Field.
    """

    name: str
    storageClass: str
    doc: str = ""
    multiple: bool = False
    deprecated: str | None = dataclasses.field(default=None, kw_only=True)

    _connection_type_set: ClassVar[str]
    _deprecation_context: str = ""

    def __post_init__(self):
        if self.deprecated and not self._deprecation_context:
            info = {}
            _ = find_outside_stacklevel("lsst.pipe.base", "dataclasses", stack_info=info)
            object.__setattr__(self, "_deprecation_context", f"{info['filename']}:{info['lineno']}")

    def __get__(self, inst, klass):
        """Descriptor access method.

        This is a method used to turn a connection into a descriptor.
        When a connection is added to a connection class, it is a class level
        variable. This method makes accessing this connection, on the
        instance of the connection class owning this connection, return a
        result specialized for that instance. In the case of connections
        this specifically means names specified in a config instance will
        be visible instead of the default names for the connection, and that
        removed connections will not be accessible on the instance.
        """
        # If inst is None, this is being accessed by the class and not an
        # instance, return this connection itself
        if inst is None:
            return self
        # Attempt to return the configured connection object from the
        # connections instance allConnections mapping.
        try:
            return inst.allConnections[self.varName]
        except KeyError:
            raise AttributeError(
                f"Connection {self.varName!r} of {klass.__name__} has been removed."
            ) from None

    # TODO: remove on DM-40443.
    @deprecated_sphinx(
        reason="Deprecated in favor of PipelineGraph, and will be removed after v27.",
        version="27.0",
        category=FutureWarning,
    )
    def makeDatasetType(
        self, universe: DimensionUniverse, parentStorageClass: StorageClass | str | None = None
    ) -> DatasetType:
        """Construct a true `~lsst.daf.butler.DatasetType` instance with
        normalized dimensions.

        Parameters
        ----------
        universe : `lsst.daf.butler.DimensionUniverse`
            Set of all known dimensions to be used to normalize the dimension
            names specified in config.
        parentStorageClass : `lsst.daf.butler.StorageClass` or `str`, optional
            Parent storage class for component datasets; `None` otherwise.

        Returns
        -------
        datasetType : `~lsst.daf.butler.DatasetType`
            The `~lsst.daf.butler.DatasetType` defined by this connection.
        """
        return DatasetType(
            self.name, universe.empty, self.storageClass, parentStorageClass=parentStorageClass
        )


@dataclasses.dataclass(frozen=True)
class DimensionedConnection(BaseConnection):
    """Class used for declaring PipelineTask connections that includes
    dimensions.

    Attributes
    ----------
    name : `str`
        The name used to identify the dataset type.
    storageClass : `str`
        The storage class used when (un)/persisting the dataset type.
    multiple : `bool`
        Indicates if this connection should expect to contain multiple objects
        of the given dataset type.  Tasks with more than one connection with
        ``multiple=True`` with the same dimensions may want to implement
        `PipelineTaskConnections.adjustQuantum` to ensure those datasets are
        consistent (i.e. zip-iterable) in `PipelineTask.runQuantum` and notify
        the execution system as early as possible of outputs that will not be
        produced because the corresponding input is missing.
    dimensions : iterable of `str`
        The `lsst.daf.butler.Butler` `lsst.daf.butler.Registry` dimensions used
        to identify the dataset type identified by the specified name.
    isCalibration : `bool`, optional
        `True` if this dataset type may be included in CALIBRATION-type
        collections to associate it with a validity range, `False` (default)
        otherwise.
    """

    dimensions: Iterable[str] = ()
    isCalibration: bool = False

    def __post_init__(self):
        super().__post_init__()
        if isinstance(self.dimensions, str):
            raise TypeError(
                "Dimensions must be iterable of dimensions, got str, possibly omitted trailing comma"
            )
        if not isinstance(self.dimensions, Iterable):
            raise TypeError("Dimensions must be iterable of dimensions")

    # TODO: remove on DM-40443.
    @deprecated_sphinx(
        reason="Deprecated in favor of PipelineGraph, and will be removed after v27.",
        version="27.0",
        category=FutureWarning,
    )
    def makeDatasetType(
        self, universe: DimensionUniverse, parentStorageClass: StorageClass | str | None = None
    ) -> DatasetType:
        """Construct a true `~lsst.daf.butler.DatasetType` instance with
        normalized dimensions.

        Parameters
        ----------
        universe : `lsst.daf.butler.DimensionUniverse`
            Set of all known dimensions to be used to normalize the dimension
            names specified in config.
        parentStorageClass : `lsst.daf.butler.StorageClass` or `str`, optional
            Parent storage class for component datasets; `None` otherwise.

        Returns
        -------
        datasetType : `~lsst.daf.butler.DatasetType`
            The `~lsst.daf.butler.DatasetType` defined by this connection.
        """
        return DatasetType(
            self.name,
            universe.conform(self.dimensions),
            self.storageClass,
            isCalibration=self.isCalibration,
            parentStorageClass=parentStorageClass,
        )


@dataclasses.dataclass(frozen=True)
class BaseInput(DimensionedConnection):
    """Class used for declaring PipelineTask input connections.

    Attributes
    ----------
    name : `str`
        The default name used to identify the dataset type.
    storageClass : `str`
        The storage class used when (un)/persisting the dataset type.
    multiple : `bool`
        Indicates if this connection should expect to contain multiple objects
        of the given dataset type.  Tasks with more than one connection with
        ``multiple=True`` with the same dimensions may want to implement
        `PipelineTaskConnections.adjustQuantum` to ensure those datasets are
        consistent (i.e. zip-iterable) in `PipelineTask.runQuantum` and notify
        the execution system as early as possible of outputs that will not be
        produced because the corresponding input is missing.
    dimensions : iterable of `str`
        The `lsst.daf.butler.Butler` `lsst.daf.butler.Registry` dimensions used
        to identify the dataset type identified by the specified name.
    deferLoad : `bool`
        Indicates that this dataset type will be loaded as a
        `lsst.daf.butler.DeferredDatasetHandle`. PipelineTasks can use this
        object to load the object at a later time.
    minimum : `bool`
        Minimum number of datasets required for this connection, per quantum.
        This is checked in the base implementation of
        `PipelineTaskConnections.adjustQuantum`, which raises `NoWorkFound` if
        the minimum is not met for `Input` connections (causing the quantum to
        be pruned, skipped, or never created, depending on the context), and
        `FileNotFoundError` for `PrerequisiteInput` connections (causing
        QuantumGraph generation to fail).  `PipelineTask` implementations may
        provide custom `~PipelineTaskConnections.adjustQuantum` implementations
        for more fine-grained or configuration-driven constraints, as long as
        they are compatible with this minium.

    Raises
    ------
    TypeError
        Raised if ``minimum`` is greater than one but ``multiple=False``.
    NotImplementedError
        Raised if ``minimum`` is zero for a regular `Input` connection; this
        is not currently supported by our QuantumGraph generation algorithm.
    """

    deferLoad: bool = False
    minimum: int = 1

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.minimum > 1 and not self.multiple:
            raise TypeError(f"Cannot set minimum={self.minimum} if multiple=False.")


@dataclasses.dataclass(frozen=True)
class Input(BaseInput):
    """Class used for declaring PipelineTask input connections.

    Attributes
    ----------
    name : `str`
        The default name used to identify the dataset type.
    storageClass : `str`
        The storage class used when (un)/persisting the dataset type.
    multiple : `bool`
        Indicates if this connection should expect to contain multiple objects
        of the given dataset type.  Tasks with more than one connection with
        ``multiple=True`` with the same dimensions may want to implement
        `PipelineTaskConnections.adjustQuantum` to ensure those datasets are
        consistent (i.e. zip-iterable) in `PipelineTask.runQuantum` and notify
        the execution system as early as possible of outputs that will not be
        produced because the corresponding input is missing.
    dimensions : iterable of `str`
        The `lsst.daf.butler.Butler` `lsst.daf.butler.Registry` dimensions used
        to identify the dataset type identified by the specified name.
    deferLoad : `bool`
        Indicates that this dataset type will be loaded as a
        `lsst.daf.butler.DeferredDatasetHandle`. PipelineTasks can use this
        object to load the object at a later time.
    minimum : `bool`
        Minimum number of datasets required for this connection, per quantum.
        This is checked in the base implementation of
        `PipelineTaskConnections.adjustQuantum`, which raises `NoWorkFound` if
        the minimum is not met for `Input` connections (causing the quantum to
        be pruned, skipped, or never created, depending on the context), and
        `FileNotFoundError` for `PrerequisiteInput` connections (causing
        QuantumGraph generation to fail).  `PipelineTask` implementations may
        provide custom `~PipelineTaskConnections.adjustQuantum` implementations
        for more fine-grained or configuration-driven constraints, as long as
        they are compatible with this minium.
    deferGraphConstraint : `bool`, optional
        If `True`, do not include this dataset type's existence in the initial
        query that starts the QuantumGraph generation process.  This can be
        used to make QuantumGraph generation faster by avoiding redundant
        datasets, and in certain cases it can (along with careful attention to
        which tasks are included in the same QuantumGraph) be used to work
        around the QuantumGraph generation algorithm's inflexible handling of
        spatial overlaps.  This option has no effect when the connection is not
        an overall input of the pipeline (or subset thereof) for which a graph
        is being created, and it never affects the ordering of quanta.
    deferBinding : `bool`, optional
        If `True`, the dataset will not be automatically included in
        the pipeline graph, ``deferGraphConstraint`` is implied.
        The custom QuantumGraphBuilder is required to bind it and add a
        corresponding edge to the pipeline graph.
        This option allows to have the same dataset type as both
        input and output of a quantum.

    Raises
    ------
    TypeError
        Raised if ``minimum`` is greater than one but ``multiple=False``.
    NotImplementedError
        Raised if ``minimum`` is zero for a regular `Input` connection; this
        is not currently supported by our QuantumGraph generation algorithm.
    """

    deferGraphConstraint: bool = False

    deferBinding: bool = False

    _connection_type_set: ClassVar[str] = "inputs"

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.minimum == 0:
            raise TypeError(f"Cannot set minimum={self.minimum} for regular input.")


@dataclasses.dataclass(frozen=True)
class PrerequisiteInput(BaseInput):
    """Class used for declaring PipelineTask prerequisite connections.

    Attributes
    ----------
    name : `str`
        The default name used to identify the dataset type.
    storageClass : `str`
        The storage class used when (un)/persisting the dataset type.
    multiple : `bool`
        Indicates if this connection should expect to contain multiple objects
        of the given dataset type.  Tasks with more than one connection with
        ``multiple=True`` with the same dimensions may want to implement
        `PipelineTaskConnections.adjustQuantum` to ensure those datasets are
        consistent (i.e. zip-iterable) in `PipelineTask.runQuantum` and notify
        the execution system as early as possible of outputs that will not be
        produced because the corresponding input is missing.
    dimensions : iterable of `str`
        The `lsst.daf.butler.Butler` `lsst.daf.butler.Registry` dimensions used
        to identify the dataset type identified by the specified name.
    minimum : `bool`
        Minimum number of datasets required for this connection, per quantum.
        This is checked in the base implementation of
        `PipelineTaskConnections.adjustQuantum`, which raises
        `FileNotFoundError` (causing QuantumGraph generation to fail).
        `PipelineTask` implementations may
        provide custom `~PipelineTaskConnections.adjustQuantum` implementations
        for more fine-grained or configuration-driven constraints, as long as
        they are compatible with this minium.
    lookupFunction : `typing.Callable`, optional
        An optional callable function that will look up PrerequisiteInputs
        using the DatasetType, registry, quantum dataId, and input collections
        passed to it. If no function is specified, the default temporal spatial
        lookup will be used.

    Raises
    ------
    TypeError
        Raised if ``minimum`` is greater than one but ``multiple=False``.

    Notes
    -----
    Prerequisite inputs are used for datasets that must exist in the data
    repository before a pipeline including this is run; they cannot be produced
    by another task in the same pipeline.

    In exchange for this limitation, they have a number of advantages relative
    to regular `Input` connections:

    - The query used to find them then during `QuantumGraph` generation can be
      fully customized by providing a ``lookupFunction``.
    - Failed searches for prerequisites during `QuantumGraph` generation will
      usually generate more helpful diagnostics than those for regular `Input`
      connections.
    - The default query for prerequisite inputs relates the quantum dimensions
      directly to the dimensions of its dataset type, without being constrained
      by any of the other dimensions in the pipeline.  This allows them to be
      used for temporal calibration lookups (which regular `Input` connections
      cannot do at present) and to work around `QuantumGraph` generation
      limitations involving cases where naive spatial overlap relationships
      between dimensions are not desired (e.g. a task that wants all detectors
      in each visit for which the visit overlaps a tract, not just those where
      that detector+visit combination overlaps the tract).
    - Prerequisite inputs may be optional (regular inputs are never optional).
    """

    lookupFunction: (
        Callable[[DatasetType, Registry, DataCoordinate, Sequence[str]], Iterable[DatasetRef]] | None
    ) = None

    _connection_type_set: ClassVar[str] = "prerequisiteInputs"


@dataclasses.dataclass(frozen=True)
class Output(DimensionedConnection):
    """Connection for output dataset."""

    _connection_type_set: ClassVar[str] = "outputs"


@dataclasses.dataclass(frozen=True)
class InitInput(BaseConnection):
    """Connection for initInput dataset."""

    _connection_type_set: ClassVar[str] = "initInputs"


@dataclasses.dataclass(frozen=True)
class InitOutput(BaseConnection):
    """Connection for initOutput dataset."""

    _connection_type_set: ClassVar[str] = "initOutputs"



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = (
    "ConnectionTypeConsistencyError",
    "DuplicateOutputError",
    "IncompatibleDatasetTypeError",
    "PipelineGraphExceptionSafetyError",
    "PipelineDataCycleError",
    "PipelineGraphError",
    "PipelineGraphReadError",
    "EdgesChangedError",
    "UnresolvedGraphError",
    "TaskNotImportedError",
)


class PipelineGraphError(RuntimeError):
    """Base exception raised when there is a problem constructing or resolving
    a pipeline graph.
    """


class DuplicateOutputError(PipelineGraphError):
    """Exception raised when multiple tasks in one pipeline produce the same
    output dataset type.
    """


class PipelineDataCycleError(PipelineGraphError):
    """Exception raised when a pipeline graph contains a cycle."""


class ConnectionTypeConsistencyError(PipelineGraphError):
    """Exception raised when the tasks in a pipeline graph use different (and
    incompatible) connection types for the same dataset type.
    """


class IncompatibleDatasetTypeError(PipelineGraphError):
    """Exception raised when the tasks in a pipeline graph define dataset types
    with the same name in incompatible ways, or when these are incompatible
    with the data repository definition.
    """


class UnresolvedGraphError(PipelineGraphError):
    """Exception raised when an operation requires dimensions or dataset types
    to have been resolved, but they have not been.
    """


class PipelineGraphReadError(PipelineGraphError, IOError):
    """Exception raised when a serialized PipelineGraph cannot be read."""


class TaskNotImportedError(PipelineGraphError):
    """Exception raised when accessing an attribute of a graph or graph node
    that is not available unless the task class has been imported and
    configured.
    """


class EdgesChangedError(PipelineGraphError):
    """Exception raised when the edges in one version of a pipeline graph
    are not consistent with those in another, but they were expected to be.
    """


class PipelineGraphExceptionSafetyError(PipelineGraphError):
    """Exception raised when a PipelineGraph method could not provide strong
    exception safety, and the graph may have been left in an inconsistent
    state.

    The originating exception is always chained when this exception is raised.
    """



#
# LSST Data Management System
# Copyright 2008, 2009, 2010 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#
import json
import logging
import numbers
import time
import unittest

import lsst.pex.config as pexConfig
import lsst.pipe.base as pipeBase
import lsst.utils.tests
import yaml

# Whilst in transition the test can't tell which type is
# going to be used for metadata.
from lsst.pipe.base.task import _TASK_METADATA_TYPE
from lsst.utils.timer import timeMethod


class AddConfig(pexConfig.Config):
    """Config for AddTask."""

    addend = pexConfig.Field(doc="amount to add", dtype=float, default=3.1)


class AddTask(pipeBase.Task):
    """Example task to add two values."""

    ConfigClass = AddConfig

    @timeMethod
    def run(self, val):
        self.metadata.add("add", self.config.addend)
        return pipeBase.Struct(
            val=val + self.config.addend,
        )


class MultConfig(pexConfig.Config):
    """Config for MultTask."""

    multiplicand = pexConfig.Field(doc="amount by which to multiply", dtype=float, default=2.5)


class MultTask(pipeBase.Task):
    """Task to multiply."""

    ConfigClass = MultConfig

    @timeMethod
    def run(self, val):
        self.metadata.add("mult", self.config.multiplicand)
        return pipeBase.Struct(
            val=val * self.config.multiplicand,
        )


# prove that registry fields can also be used to hold subtasks
# by using a registry to hold MultTask
multRegistry = pexConfig.makeRegistry("Registry for Mult-like tasks")
multRegistry.register("stdMult", MultTask)


class AddMultConfig(pexConfig.Config):
    """Config for AddMult."""

    add = AddTask.makeField("add task")
    mult = multRegistry.makeField("mult task", default="stdMult")


class AddMultTask(pipeBase.Task):
    """Test Task with subtasks."""

    ConfigClass = AddMultConfig
    _DefaultName = "addMult"
    _add_module_logger_prefix = False

    """First add, then multiply."""

    def __init__(self, **keyArgs):
        pipeBase.Task.__init__(self, **keyArgs)
        self.makeSubtask("add")
        self.makeSubtask("mult")

    @timeMethod
    def run(self, val):
        with self.timer("context"):
            addRet = self.add.run(val)
            multRet = self.mult.run(addRet.val)
            self.metadata.add("addmult", multRet.val)
            return pipeBase.Struct(
                val=multRet.val,
            )

    @timeMethod
    def failDec(self):
        """Fail with a decorator."""
        raise RuntimeError("failDec intentional error")

    def failCtx(self):
        """Fail inside a context manager."""
        with self.timer("failCtx"):
            raise RuntimeError("failCtx intentional error")


class AddMultTask2(AddMultTask):
    """Subclass that gets an automatic logger prefix."""

    _add_module_logger_prefix = True


class AddTwiceTask(AddTask):
    """Variant of AddTask that adds twice the addend."""

    def run(self, val):
        addend = self.config.addend
        return pipeBase.Struct(val=val + (2 * addend))


class TaskTestCase(unittest.TestCase):
    """A test case for Task."""

    def testBasics(self):
        """Test basic construction and use of a task."""
        for addend in (1.1, -3.5):
            for multiplicand in (0.9, -45.0):
                config = AddMultTask.ConfigClass()
                config.add.addend = addend
                config.mult["stdMult"].multiplicand = multiplicand
                # make sure both ways of accessing the registry work and give
                # the same result
                self.assertEqual(config.mult.active.multiplicand, multiplicand)
                addMultTask = AddMultTask(config=config)
                for val in (-1.0, 0.0, 17.5):
                    ret = addMultTask.run(val=val)
                    self.assertAlmostEqual(ret.val, (val + addend) * multiplicand)

    def testNames(self):
        """Test getName() and getFullName()."""
        addMultTask = AddMultTask()
        self.assertEqual(addMultTask.getName(), "addMult")
        self.assertEqual(addMultTask.add.getName(), "add")
        self.assertEqual(addMultTask.mult.getName(), "mult")

        self.assertEqual(addMultTask._name, "addMult")
        self.assertEqual(addMultTask.add._name, "add")
        self.assertEqual(addMultTask.mult._name, "mult")

        self.assertEqual(addMultTask.getFullName(), "addMult")
        self.assertEqual(addMultTask.add.getFullName(), "addMult.add")
        self.assertEqual(addMultTask.mult.getFullName(), "addMult.mult")

        self.assertEqual(addMultTask._fullName, "addMult")
        self.assertEqual(addMultTask.add._fullName, "addMult.add")
        self.assertEqual(addMultTask.mult._fullName, "addMult.mult")

    def testLog(self):
        """Test the Task's logger."""
        addMultTask = AddMultTask()
        self.assertEqual(addMultTask.log.name, "addMult")
        self.assertEqual(addMultTask.add.log.name, "addMult.add")

        log = logging.getLogger("tester")
        addMultTask = AddMultTask(log=log)
        self.assertEqual(addMultTask.log.name, "tester.addMult")
        self.assertEqual(addMultTask.add.log.name, "tester.addMult.add")

        addMultTask2 = AddMultTask2()
        self.assertEqual(addMultTask2.log.name, f"{__name__}.addMult")

    def testGetFullMetadata(self):
        """Test getFullMetadata()."""
        addMultTask = AddMultTask()
        addMultTask.run(val=1.234)  # Add some metadata
        fullMetadata = addMultTask.getFullMetadata()
        self.assertIsInstance(fullMetadata["addMult"], _TASK_METADATA_TYPE)
        self.assertIsInstance(fullMetadata["addMult:add"], _TASK_METADATA_TYPE)
        self.assertIsInstance(fullMetadata["addMult:mult"], _TASK_METADATA_TYPE)
        self.assertEqual(set(fullMetadata), {"addMult", "addMult:add", "addMult:mult"})

        all_names = fullMetadata.names()
        self.assertIn("addMult", all_names)
        self.assertIn("addMult.runStartUtc", all_names)

        param_names = fullMetadata.paramNames(topLevelOnly=True)
        # No top level keys without hierarchy
        self.assertEqual(set(param_names), set())

        param_names = fullMetadata.paramNames(topLevelOnly=False)
        self.assertNotIn("addMult", param_names)
        self.assertIn("addMult.runStartUtc", param_names)
        self.assertIn("addMult:add.runStartCpuTime", param_names)

    def testEmptyMetadata(self):
        task = AddMultTask()
        task.run(val=1.2345)
        task.emptyMetadata()
        fullMetadata = task.getFullMetadata()
        self.assertEqual(len(fullMetadata["addMult"]), 0)
        self.assertEqual(len(fullMetadata["addMult:add"]), 0)
        self.assertEqual(len(fullMetadata["addMult:mult"]), 0)

    def testReplace(self):
        """Test replacing one subtask with another."""
        for addend in (1.1, -3.5):
            for multiplicand in (0.9, -45.0):
                config = AddMultTask.ConfigClass()
                config.add.retarget(AddTwiceTask)
                config.add.addend = addend
                config.mult["stdMult"].multiplicand = multiplicand
                addMultTask = AddMultTask(config=config)
                for val in (-1.0, 0.0, 17.5):
                    ret = addMultTask.run(val=val)
                    self.assertAlmostEqual(ret.val, (val + (2 * addend)) * multiplicand)

    def testFail(self):
        """Test timers when the code they are timing fails."""
        addMultTask = AddMultTask()
        try:
            addMultTask.failDec()
            self.fail("Expected RuntimeError")
        except RuntimeError:
            self.assertIn("failDecEndCpuTime", addMultTask.metadata)
        try:
            addMultTask.failCtx()
            self.fail("Expected RuntimeError")
        except RuntimeError:
            self.assertIn("failCtxEndCpuTime", addMultTask.metadata)

    def testTimeMethod(self):
        """Test that the timer is adding the right metadata."""
        addMultTask = AddMultTask()

        # Run twice to ensure we are additive.
        addMultTask.run(val=1.1)
        addMultTask.run(val=2.0)
        # Check existence and type
        for key, keyType in (
            ("Utc", str),
            ("CpuTime", float),
            ("UserTime", float),
            ("SystemTime", float),
            ("MaxResidentSetSize", numbers.Integral),
            ("MinorPageFaults", numbers.Integral),
            ("MajorPageFaults", numbers.Integral),
            ("BlockInputs", numbers.Integral),
            ("BlockOutputs", numbers.Integral),
            ("VoluntaryContextSwitches", numbers.Integral),
            ("InvoluntaryContextSwitches", numbers.Integral),
        ):
            for when in ("Start", "End"):
                for method in ("run", "context"):
                    name = method + when + key
                    self.assertIn(name, addMultTask.metadata, name + " is missing from task metadata")
                    self.assertIsInstance(
                        addMultTask.metadata.getScalar(name),
                        keyType,
                        f"{name} is not of the right type "
                        f"({keyType} vs {type(addMultTask.metadata.getScalar(name))})",
                    )
        # Some basic sanity checks
        currCpuTime = time.process_time()
        self.assertLessEqual(
            addMultTask.metadata.getScalar("runStartCpuTime"),
            addMultTask.metadata.getScalar("runEndCpuTime"),
        )
        self.assertLessEqual(addMultTask.metadata.getScalar("runEndCpuTime"), currCpuTime)
        self.assertLessEqual(
            addMultTask.metadata.getScalar("contextStartCpuTime"),
            addMultTask.metadata.getScalar("contextEndCpuTime"),
        )
        self.assertLessEqual(addMultTask.metadata.getScalar("contextEndCpuTime"), currCpuTime)
        self.assertLessEqual(
            addMultTask.add.metadata.getScalar("runStartCpuTime"),
            addMultTask.metadata.getScalar("runEndCpuTime"),
        )
        self.assertLessEqual(addMultTask.add.metadata.getScalar("runEndCpuTime"), currCpuTime)

        # Add some explicit values for serialization test.
        addMultTask.metadata["comment"] = "A comment"
        addMultTask.metadata["integer"] = 5
        addMultTask.metadata["float"] = 3.14
        addMultTask.metadata["bool"] = False
        addMultTask.metadata.add("commentList", "comment1")
        addMultTask.metadata.add("commentList", "comment1")
        addMultTask.metadata.add("intList", 6)
        addMultTask.metadata.add("intList", 7)
        addMultTask.metadata.add("boolList", False)
        addMultTask.metadata.add("boolList", True)
        addMultTask.metadata.add("floatList", 6.6)
        addMultTask.metadata.add("floatList", 7.8)

        # TaskMetadata can serialize to JSON but not YAML
        # and PropertySet can serialize to YAML and not JSON.
        if hasattr(addMultTask.metadata, "json"):
            j = addMultTask.metadata.model_dump_json()
            new_meta = pipeBase.TaskMetadata.model_validate(json.loads(j))
        else:
            y = yaml.dump(addMultTask.metadata)
            new_meta = yaml.safe_load(y)
        self.assertEqual(new_meta, addMultTask.metadata)

    def test_annotate_exception(self):
        """Test annotating failures in the task metadata when a non-Task
        exception is raised (when there is no `metadata` on the exception).
        """
        task = AddMultTask()
        msg = "something failed!"
        error = ValueError(msg)
        with self.assertLogs("addMult", level="ERROR") as cm:
            pipeBase.AnnotatedPartialOutputsError.annotate(error, task, log=task.log)
        self.assertIn(msg, "\n".join(cm.output))
        self.assertEqual(task.metadata["failure"]["message"], msg)
        self.assertEqual(task.metadata["failure"]["type"], "ValueError")
        self.assertNotIn("metadata", task.metadata["failure"])

    def test_annotate_task_exception(self):
        """Test annotating failures in the task metadata when a Task-specific
        exception is raised.
        """

        class TestError(pipeBase.AlgorithmError):
            @property
            def metadata(self):
                return {"something": 12345}

        task = AddMultTask()
        msg = "something failed!"
        error = TestError(msg)
        with self.assertLogs("addMult", level="ERROR") as cm:
            pipeBase.AnnotatedPartialOutputsError.annotate(error, task, log=task.log)
        self.assertIn(msg, "\n".join(cm.output))
        self.assertEqual(task.metadata["failure"]["message"], msg)
        result = "test_task.TaskTestCase.test_annotate_task_exception.<locals>.TestError"
        self.assertEqual(task.metadata["failure"]["type"], result)
        self.assertEqual(task.metadata["failure"]["metadata"]["something"], 12345)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Run file leak tests."""


def setup_module(module):
    """Configure pytest."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping, Sequence
from typing import Any, ClassVar, TypeVar, cast, overload

import networkx

from ._dataset_types import DatasetTypeNode
from ._exceptions import UnresolvedGraphError
from ._nodes import NodeKey, NodeType
from ._tasks import TaskInitNode, TaskNode

_N = TypeVar("_N", covariant=True)
_T = TypeVar("_T")


class MappingView(Mapping[str, _N]):
    """Base class for mapping views into nodes of certain types in a
    `PipelineGraph`.

    Parameters
    ----------
    parent_xgraph : `networkx.MultiDiGraph`
        Backing networkx graph for the `PipelineGraph` instance.

    Notes
    -----
    Instances should only be constructed by `PipelineGraph` and its helper
    classes.

    Iteration order is topologically sorted if and only if the backing
    `PipelineGraph` has been sorted since its last modification.
    """

    def __init__(self, parent_xgraph: networkx.MultiDiGraph) -> None:
        self._parent_xgraph = parent_xgraph
        self._keys: list[str] | None = None

    _NODE_TYPE: ClassVar[NodeType]  # defined by derived classes

    def __contains__(self, key: object) -> bool:
        # The given key may not be a str, but if it isn't it'll just fail the
        # check, which is what we want anyway.
        return NodeKey(self._NODE_TYPE, cast(str, key)) in self._parent_xgraph

    def __iter__(self) -> Iterator[str]:
        if self._keys is None:
            self._keys = self._make_keys(self._parent_xgraph)
        return iter(self._keys)

    def __getitem__(self, key: str) -> _N:
        return self._parent_xgraph.nodes[NodeKey(self._NODE_TYPE, key)]["instance"]

    def __len__(self) -> int:
        if self._keys is None:
            self._keys = self._make_keys(self._parent_xgraph)
        return len(self._keys)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self!s})"

    def __str__(self) -> str:
        return f"{{{', '.join(iter(self))}}}"

    def _reorder(self, parent_keys: Sequence[NodeKey]) -> None:
        """Set this view's iteration order according to the given iterable of
        parent keys.

        Parameters
        ----------
        parent_keys : `~collections.abc.Sequence` [ `NodeKey` ]
            Superset of the keys in this view, in the new order.
        """
        self._keys = self._make_keys(parent_keys)

    def _reset(self) -> None:
        """Reset all cached content.

        This should be called by the parent graph after any changes that could
        invalidate the view, causing it to be reconstructed when next
        requested.
        """
        self._keys = None

    def _make_keys(self, parent_keys: Iterable[NodeKey]) -> list[str]:
        """Make a sequence of keys for this view from an iterable of parent
        keys.

        Parameters
        ----------
        parent_keys : `~collections.abc.Iterable` [ `NodeKey` ]
            Superset of the keys in this view.
        """
        return [str(k) for k in parent_keys if k.node_type is self._NODE_TYPE]


class TaskMappingView(MappingView[TaskNode]):
    """A mapping view of the tasks in a `PipelineGraph`.

    Notes
    -----
    Mapping keys are task labels and values are `TaskNode` instances.
    Iteration order is topological if and only if the `PipelineGraph` has been
    sorted since its last modification.
    """

    _NODE_TYPE = NodeType.TASK

    def between(self, first: str | None = None, last: str | None = None) -> Mapping[str, TaskNode]:
        """Return a mapping whose tasks are between a range of tasks.

        Parameters
        ----------
        first : `str`, optional
            Label of the first task to include, inclusive.
        last : `str`, optional
            Label of the last task to include, inclusive.

        Returns
        -------
        between : `~collections.abc.Mapping` [ `str`, `TaskNode` ]
            Tasks that are downstream of ``first`` and upstream of ``last``.
            If either ``first`` or ``last`` is `None` (default), that side
            of the result is unbounded.  Tasks that have no dependency
            relationship to either task are not included, and if both bounds
            are provided, included tasks must have the right relationship with
            both bounding tasks.
        """
        # This is definitely not the fastest way to compute this subset, but
        # it's a very simple one (given what networkx provides), and pipeline
        # graphs are never *that* big.
        if first is not None:
            first_key = NodeKey(NodeType.TASK, first)
            a: set[NodeKey] = set(networkx.dag.descendants(self._parent_xgraph, first_key))
            a.add(first_key)
        else:
            a = set(self._parent_xgraph.nodes.keys())
        if last is not None:
            last_key = NodeKey(NodeType.TASK, last)
            b: set[NodeKey] = set(networkx.dag.ancestors(self._parent_xgraph, last_key))
            b.add(last_key)
        else:
            b = set(self._parent_xgraph.nodes.keys())
        return {
            key.name: self._parent_xgraph.nodes[key]["instance"]
            for key in a & b
            if key.node_type is NodeType.TASK
        }


class TaskInitMappingView(MappingView[TaskInitNode]):
    """A mapping view of the nodes representing task initialization in a
    `PipelineGraph`.

    Notes
    -----
    Mapping keys are task labels and values are `TaskInitNode` instances.
    Iteration order is topological if and only if the `PipelineGraph` has been
    sorted since its last modification.
    """

    _NODE_TYPE = NodeType.TASK_INIT


class DatasetTypeMappingView(MappingView[DatasetTypeNode]):
    """A mapping view of the nodes representing task initialization in a
    `PipelineGraph`.

    Notes
    -----
    Mapping keys are parent dataset type names and values are `DatasetTypeNode`
    instances, but values are only available for nodes that have been resolved
    (see `PipelineGraph.resolve`).  Attempting to access an unresolved value
    will result in `UnresolvedGraphError` being raised.  Keys for unresolved
    nodes are always present and iterable.

    Iteration order is topological if and only if the `PipelineGraph` has been
    sorted since its last modification.
    """

    _NODE_TYPE = NodeType.DATASET_TYPE

    def __getitem__(self, key: str) -> DatasetTypeNode:
        if (result := super().__getitem__(key)) is None:
            raise UnresolvedGraphError(f"Node for dataset type {key!r} has not been resolved.")
        return result

    def is_resolved(self, key: str) -> bool:
        """Test whether a node has been resolved.

        Parameters
        ----------
        key : `str`
            Node to check.

        Returns
        -------
        `bool`
            Whether the node has been resolved or not.
        """
        return super().__getitem__(key) is not None

    @overload
    def get_if_resolved(self, key: str) -> DatasetTypeNode | None: ...  # pragma: nocover

    @overload
    def get_if_resolved(self, key: str, default: _T) -> DatasetTypeNode | _T: ...  # pragma: nocover

    def get_if_resolved(self, key: str, default: Any = None) -> DatasetTypeNode | Any:
        """Get a node or return a default if it has not been resolved.

        Parameters
        ----------
        key : `str`
            Parent dataset type name.
        default : `~typing.Any`
            Value to return if this dataset type has not been resolved.

        Returns
        -------
        result : `DatasetTypeNode`
            The resolved node or the default value.

        Raises
        ------
        KeyError
            Raised if the node is not present in the graph at all.
        """
        if (result := super().__getitem__(key)) is None:
            return default  # type: ignore
        return result



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining variants for valid values used to constrain datasets in a
graph building query.
"""

from __future__ import annotations

__all__ = ("ExecutionResources", "QuantumContext")

import numbers
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import Any

import astropy.units as u
from lsst.daf.butler import DatasetRef, DimensionUniverse, LimitedButler, Quantum
from lsst.utils.introspection import get_full_type_name
from lsst.utils.logging import PeriodicLogger, getLogger

from .connections import DeferredDatasetRef, InputQuantizedConnection, OutputQuantizedConnection
from .struct import Struct

_LOG = getLogger(__name__)


@dataclass(init=False, frozen=True)
class ExecutionResources:
    """A description of the resources available to a running quantum.

    Parameters
    ----------
    num_cores : `int`, optional
        The number of cores allocated to the task.
    max_mem : `~astropy.units.Quantity`, `numbers.Real`, `str`, or `None`,\
            optional
        The amount of memory allocated to the task. Can be specified
        as byte-compatible `~astropy.units.Quantity`, a plain number,
        a string with a plain number, or a string representing a quantity.
        If `None` no limit is specified.
    default_mem_units : `astropy.units.Unit`, optional
        The default unit to apply when the ``max_mem`` value is given
        as a plain number.
    """

    num_cores: int = 1
    """The maximum number of cores that the task can use."""

    max_mem: u.Quantity | None = None
    """If defined, the amount of memory allocated to the task.
    """

    def __init__(
        self,
        *,
        num_cores: int = 1,
        max_mem: u.Quantity | numbers.Real | str | None = None,
        default_mem_units: u.Unit = u.B,
    ):
        # Create our own __init__ to allow more flexible input parameters
        # but with a constrained dataclass definition.
        if num_cores < 1:
            raise ValueError("The number of cores must be a positive integer")

        object.__setattr__(self, "num_cores", num_cores)

        mem: u.Quantity | None = None

        if max_mem is None or isinstance(max_mem, u.Quantity):
            mem = max_mem
        elif max_mem == "":
            # Some command line tooling can treat no value as empty string.
            pass
        else:
            parsed_mem = None
            try:
                parsed_mem = float(max_mem)
            except ValueError:
                pass
            else:
                mem = parsed_mem * default_mem_units

            if mem is None:
                mem = u.Quantity(max_mem)

        if mem is not None:
            # Force to bytes. This also checks that we can convert to bytes.
            mem = mem.to(u.B)

        object.__setattr__(self, "max_mem", mem)

    def __deepcopy__(self, memo: Any) -> ExecutionResources:
        """Deep copy returns itself because the class is frozen."""
        return self

    def _reduce_kwargs(self) -> dict[str, Any]:
        """Return a dict of the keyword arguments that should be used
        by `__reduce__`.

        This is necessary because the dataclass is defined to be keyword
        only and we wish the default pickling to only store a plain number
        for the memory allocation and not a large Quantity.

        Returns
        -------
        kwargs : `dict`
            Keyword arguments to be used when pickling.
        """
        kwargs: dict[str, Any] = {"num_cores": self.num_cores}
        if self.max_mem is not None:
            # .value is a numpy float. Cast it to a python int since we
            # do not want fractional bytes. The constructor ensures that this
            # uses units of byte so we do not have to convert.
            kwargs["max_mem"] = int(self.max_mem.value)
        return kwargs

    @staticmethod
    def _unpickle_via_factory(
        cls: type[ExecutionResources], args: Sequence[Any], kwargs: dict[str, Any]
    ) -> ExecutionResources:
        """Unpickle something by calling a factory.

        Allows unpickle using `__reduce__` with keyword
        arguments as well as positional arguments.
        """
        return cls(**kwargs)

    def __reduce__(
        self,
    ) -> tuple[
        Callable[[type[ExecutionResources], Sequence[Any], dict[str, Any]], ExecutionResources],
        tuple[type[ExecutionResources], Sequence[Any], dict[str, Any]],
    ]:
        """Pickler."""
        return self._unpickle_via_factory, (self.__class__, [], self._reduce_kwargs())


class QuantumContext:
    """A Butler-like class specialized for a single quantum along with
    context information that can influence how the task is executed.

    Parameters
    ----------
    butler : `lsst.daf.butler.LimitedButler`
        Butler object from/to which datasets will be get/put.
    quantum : `lsst.daf.butler.Quantum`
        Quantum object that describes the datasets which will be get/put by a
        single execution of this node in the pipeline graph.
    resources : `ExecutionResources`, optional
        The resources allocated for executing quanta.

    Notes
    -----
    A `QuantumContext` class wraps a standard butler interface and
    specializes it to the context of a given quantum. What this means
    in practice is that the only gets and puts that this class allows
    are DatasetRefs that are contained in the quantum.

    In the future this class will also be used to record provenance on
    what was actually get and put. This is in contrast to what the
    preflight expects to be get and put by looking at the graph before
    execution.
    """

    resources: ExecutionResources

    def __init__(
        self, butler: LimitedButler, quantum: Quantum, *, resources: ExecutionResources | None = None
    ):
        self.quantum = quantum
        if resources is None:
            resources = ExecutionResources()
        self.resources = resources

        self.allInputs = set()
        self.allOutputs = set()
        for refs in quantum.inputs.values():
            for ref in refs:
                self.allInputs.add((ref.datasetType, ref.dataId))
        for refs in quantum.outputs.values():
            for ref in refs:
                self.allOutputs.add((ref.datasetType, ref.dataId))
        self.__butler = butler

    def _get(self, ref: DeferredDatasetRef | DatasetRef | None) -> Any:
        # Butler methods below will check for unresolved DatasetRefs and
        # raise appropriately, so no need for us to do that here.
        if isinstance(ref, DeferredDatasetRef):
            self._checkMembership(ref.datasetRef, self.allInputs)
            return self.__butler.getDeferred(ref.datasetRef)
        elif ref is None:
            return None
        else:
            self._checkMembership(ref, self.allInputs)
            return self.__butler.get(ref)

    def _put(self, value: Any, ref: DatasetRef) -> None:
        """Store data in butler."""
        self._checkMembership(ref, self.allOutputs)
        self.__butler.put(value, ref)

    def get(
        self,
        dataset: (
            InputQuantizedConnection
            | list[DatasetRef | None]
            | list[DeferredDatasetRef | None]
            | DatasetRef
            | DeferredDatasetRef
            | None
        ),
    ) -> Any:
        """Fetch data from the butler.

        Parameters
        ----------
        dataset : see description
            This argument may either be an `InputQuantizedConnection` which
            describes all the inputs of a quantum, a list of
            `~lsst.daf.butler.DatasetRef`, or a single
            `~lsst.daf.butler.DatasetRef`. The function will get and return
            the corresponding datasets from the butler. If `None` is passed in
            place of a `~lsst.daf.butler.DatasetRef` then the corresponding
            returned object will be `None`.

        Returns
        -------
        return : `object`
            This function returns arbitrary objects fetched from the bulter.
            The structure these objects are returned in depends on the type of
            the input argument. If the input dataset argument is a
            `InputQuantizedConnection`, then the return type will be a
            dictionary with keys corresponding to the attributes of the
            `InputQuantizedConnection` (which in turn are the attribute
            identifiers of the connections). If the input argument is of type
            `list` of `~lsst.daf.butler.DatasetRef` then the return type will
            be a list of objects.  If the input argument is a single
            `~lsst.daf.butler.DatasetRef` then a single object will be
            returned.

        Raises
        ------
        ValueError
            Raised if a `~lsst.daf.butler.DatasetRef` is passed to get that is
            not defined in the quantum object
        """
        # Set up a periodic logger so log messages can be issued if things
        # are taking too long.
        periodic = PeriodicLogger(_LOG)

        if isinstance(dataset, InputQuantizedConnection):
            retVal = {}
            n_connections = len(dataset)
            n_retrieved = 0
            for i, (name, ref) in enumerate(dataset):
                if isinstance(ref, list | tuple):
                    val = []
                    n_refs = len(ref)
                    for j, r in enumerate(ref):
                        val.append(self._get(r))
                        n_retrieved += 1
                        periodic.log(
                            "Retrieved %d out of %d datasets for connection '%s' (%d out of %d)",
                            j + 1,
                            n_refs,
                            name,
                            i + 1,
                            n_connections,
                        )
                else:
                    val = self._get(ref)
                    periodic.log(
                        "Retrieved dataset for connection '%s' (%d out of %d)",
                        name,
                        i + 1,
                        n_connections,
                    )
                    n_retrieved += 1
                retVal[name] = val
            if periodic.num_issued > 0:
                # This took long enough that we issued some periodic log
                # messages, so issue a final confirmation message as well.
                _LOG.verbose(
                    "Completed retrieval of %d datasets from %d connections", n_retrieved, n_connections
                )
            return retVal
        elif isinstance(dataset, list | tuple):
            n_datasets = len(dataset)
            retrieved = []
            for i, x in enumerate(dataset):
                # Mypy is not sure of the type of x because of the union
                # of lists so complains. Ignoring it is more efficient
                # than adding an isinstance assert.
                retrieved.append(self._get(x))
                periodic.log("Retrieved %d out of %d datasets", i + 1, n_datasets)
            if periodic.num_issued > 0:
                _LOG.verbose("Completed retrieval of %d datasets", n_datasets)
            return retrieved
        elif isinstance(dataset, DatasetRef | DeferredDatasetRef) or dataset is None:
            return self._get(dataset)
        else:
            raise TypeError(
                f"Dataset argument ({get_full_type_name(dataset)}) is not a type that can be used to get"
            )

    def put(
        self,
        values: Struct | list[Any] | Any,
        dataset: OutputQuantizedConnection | list[DatasetRef] | DatasetRef,
    ) -> None:
        """Put data into the butler.

        Parameters
        ----------
        values : `Struct` or `list` of `object` or `object`
            The data that should be put with the butler. If the type of the
            dataset is `OutputQuantizedConnection` then this argument should be
            a `Struct` with corresponding attribute names. Each attribute
            should then correspond to either a list of object or a single
            object depending of the type of the corresponding attribute on
            dataset. I.e. if ``dataset.calexp`` is
            ``[datasetRef1, datasetRef2]`` then ``values.calexp`` should be
            ``[calexp1, calexp2]``. Like wise if there is a single ref, then
            only a single object need be passed. The same restriction applies
            if dataset is directly a `list` of `~lsst.daf.butler.DatasetRef`
            or a single `~lsst.daf.butler.DatasetRef`. If ``values.NAME`` is
            None, no output is written.
        dataset : `OutputQuantizedConnection` or `list`[`DatasetRef`] \
                or `DatasetRef`
            This argument may either be an `InputQuantizedConnection` which
            describes all the inputs of a quantum, a list of
            `lsst.daf.butler.DatasetRef`, or a single
            `lsst.daf.butler.DatasetRef`. The function will get and return
            the corresponding datasets from the butler.

        Raises
        ------
        ValueError
            Raised if a `~lsst.daf.butler.DatasetRef` is passed to put that is
            not defined in the `~lsst.daf.butler.Quantum` object, or the type
            of values does not match what is expected from the type of dataset.
        """
        if isinstance(dataset, OutputQuantizedConnection):
            if not isinstance(values, Struct):
                raise ValueError(
                    "dataset is a OutputQuantizedConnection, a Struct with corresponding"
                    " attributes must be passed as the values to put"
                )
            for name, refs in dataset:
                if (valuesAttribute := getattr(values, name, None)) is None:
                    continue
                if isinstance(refs, list | tuple):
                    if len(refs) != len(valuesAttribute):
                        raise ValueError(f"There must be a object to put for every Dataset ref in {name}")
                    for i, ref in enumerate(refs):
                        self._put(valuesAttribute[i], ref)
                else:
                    self._put(valuesAttribute, refs)
        elif isinstance(dataset, list | tuple):
            if not isinstance(values, Sequence):
                raise ValueError("Values to put must be a sequence")
            if len(dataset) != len(values):
                raise ValueError("There must be a common number of references and values to put")
            for i, ref in enumerate(dataset):
                self._put(values[i], ref)
        elif isinstance(dataset, DatasetRef):
            self._put(values, dataset)
        else:
            raise TypeError("Dataset argument is not a type that can be used to put")

    def _checkMembership(self, ref: list[DatasetRef] | DatasetRef, inout: set) -> None:
        """Check if a `~lsst.daf.butler.DatasetRef` is part of the input
        `~lsst.daf.butler.Quantum`.

        This function will raise an exception if the `QuantumContext` is
        used to get/put a `~lsst.daf.butler.DatasetRef` which is not defined
        in the quantum.

        Parameters
        ----------
        ref : `list` [ `~lsst.daf.butler.DatasetRef` ] or \
                `~lsst.daf.butler.DatasetRef`
            Either a `list` or a single `~lsst.daf.butler.DatasetRef` to check
        inout : `set`
            The connection type to check, e.g. either an input or an output.
            This prevents both types needing to be checked for every operation,
            which may be important for Quanta with lots of
            `~lsst.daf.butler.DatasetRef`.
        """
        if not isinstance(ref, list | tuple):
            ref = [ref]
        for r in ref:
            if (r.datasetType, r.dataId) not in inout:
                raise ValueError("DatasetRef is not part of the Quantum being processed")

    @property
    def dimensions(self) -> DimensionUniverse:
        """Structure managing all dimensions recognized by this data
        repository (`~lsst.daf.butler.DimensionUniverse`).
        """
        return self.__butler.dimensions



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for PipelineTask.
"""

import copy
import pickle
import unittest
from typing import Any

import astropy.units as u
import lsst.pex.config as pexConfig
import lsst.pipe.base as pipeBase
import lsst.utils.logging
import lsst.utils.tests
from lsst.daf.butler import DataCoordinate, DatasetRef, DatasetType, DimensionUniverse, Quantum


class ButlerMock:
    """Mock version of butler, only usable for this test."""

    def __init__(self) -> None:
        self.datasets: dict[str, dict[DataCoordinate, Any]] = {}
        self.dimensions = DimensionUniverse()

    def get(self, ref: DatasetRef) -> Any:
        dsdata = self.datasets.get(ref.datasetType.name)
        if dsdata:
            return dsdata.get(ref.dataId)
        return None

    def put(self, inMemoryDataset: Any, dsRef: DatasetRef, producer: Any = None):
        key = dsRef.dataId
        name = dsRef.datasetType.name
        dsdata = self.datasets.setdefault(name, {})
        dsdata[key] = inMemoryDataset


class AddConnections(pipeBase.PipelineTaskConnections, dimensions=["instrument", "visit"]):
    """Connections for the AddTask."""

    input = pipeBase.connectionTypes.Input(
        name="add_input",
        dimensions=["instrument", "visit", "detector", "physical_filter", "band"],
        storageClass="Catalog",
        doc="Input dataset type for this task",
    )
    output = pipeBase.connectionTypes.Output(
        name="add_output",
        dimensions=["instrument", "visit", "detector", "physical_filter", "band"],
        storageClass="Catalog",
        doc="Output dataset type for this task",
    )


class AddConfig(pipeBase.PipelineTaskConfig, pipelineConnections=AddConnections):
    """Config for the AddTask."""

    addend = pexConfig.Field[int](doc="amount to add", default=3)


class AddTask(pipeBase.PipelineTask):
    """Example task which overrides run() method."""

    ConfigClass = AddConfig
    _DefaultName = "add_task"

    def run(self, input: int) -> pipeBase.Struct:
        self.metadata.add("add", self.config.addend)
        output = input + self.config.addend
        return pipeBase.Struct(output=output)


class AddTask2(pipeBase.PipelineTask):
    """Example task which overrides runQuantum() method."""

    ConfigClass = AddConfig
    _DefaultName = "add_task_2"

    def runQuantum(
        self, butlerQC: pipeBase.QuantumContext, inputRefs: DatasetRef, outputRefs: DatasetRef
    ) -> None:
        self.metadata.add("add", self.config.addend)
        inputs = butlerQC.get(inputRefs)
        outputs = inputs["input"] + self.config.addend
        butlerQC.put(pipeBase.Struct(output=outputs), outputRefs)


class PipelineTaskTestCase(unittest.TestCase):
    """A test case for PipelineTask."""

    def _makeDSRefVisit(self, dstype: DatasetType, visitId: int) -> DatasetRef:
        dataId = DataCoordinate.standardize(
            detector="X",
            visit=visitId,
            physical_filter="a",
            band="b",
            instrument="TestInstrument",
            graph=dstype.dimensions,
        )
        run = "test"
        ref = DatasetRef(datasetType=dstype, dataId=dataId, run=run)
        return ref

    def _makeQuanta(
        self,
        task_node: pipeBase.pipeline_graph.TaskNode,
        pipeline_graph: pipeBase.PipelineGraph,
        nquanta: int = 100,
    ) -> list[Quantum]:
        """Create set of Quanta"""
        quanta = []
        for visit in range(nquanta):
            quantum = Quantum(
                inputs={
                    pipeline_graph.dataset_types[edge.dataset_type_name].dataset_type: [
                        self._makeDSRefVisit(
                            pipeline_graph.dataset_types[edge.dataset_type_name].dataset_type, visit
                        )
                    ]
                    for edge in task_node.inputs.values()
                },
                outputs={
                    pipeline_graph.dataset_types[edge.dataset_type_name].dataset_type: [
                        self._makeDSRefVisit(
                            pipeline_graph.dataset_types[edge.dataset_type_name].dataset_type, visit
                        )
                    ]
                    for edge in task_node.outputs.values()
                },
            )
            quanta.append(quantum)

        return quanta

    def testRunQuantumFull(self):
        """Test for AddTask.runQuantum() implementation with full butler."""
        self._testRunQuantum(full_butler=True)

    def testRunQuantumLimited(self):
        """Test for AddTask.runQuantum() implementation with limited butler."""
        self._testRunQuantum(full_butler=False)

    def _testRunQuantum(self, full_butler: bool) -> None:
        """Test for AddTask.runQuantum() implementation."""
        butler = ButlerMock()
        task = AddTask(config=AddConfig())
        input_name = task.config.connections.input
        output_name = task.config.connections.output

        pipeline_graph = pipeBase.PipelineGraph()
        task_node = pipeline_graph.add_task(None, type(task), task.config)
        pipeline_graph.resolve(dimensions=butler.dimensions, dataset_types={})

        # make all quanta
        quanta = self._makeQuanta(task_node, pipeline_graph)

        # add input data to butler
        dstype0 = pipeline_graph.dataset_types[input_name].dataset_type
        for i, quantum in enumerate(quanta):
            ref = quantum.inputs[dstype0.name][0]
            butler.put(100 + i, ref)

        # run task on each quanta
        checked_get = False
        for quantum in quanta:
            butlerQC = pipeBase.QuantumContext(butler, quantum)
            inputRefs, outputRefs = task_node.get_connections().buildDatasetRefs(quantum)
            task.runQuantum(butlerQC, inputRefs, outputRefs)

            # Test getting of datasets in different ways.
            # (only need to do this one time)
            if not checked_get:
                # Force the periodic logger to issue messages.
                lsst.utils.logging.PeriodicLogger.LOGGING_INTERVAL = 0.0

                checked_get = True
                with self.assertLogs("lsst.pipe.base", level=lsst.utils.logging.VERBOSE) as cm:
                    input_data = butlerQC.get(inputRefs)
                self.assertIn("Completed", cm.output[-1])
                self.assertEqual(len(input_data), len(inputRefs))

                # In this test there are no multiples returned.
                refs = [ref for _, ref in inputRefs]
                with self.assertLogs("lsst.pipe.base", level=lsst.utils.logging.VERBOSE) as cm:
                    list_get = butlerQC.get(refs)

                lsst.utils.logging.PeriodicLogger.LOGGING_INTERVAL = 600.0

                self.assertIn("Completed", cm.output[-1])
                self.assertEqual(len(list_get), len(input_data))
                self.assertIsInstance(list_get[0], int)
                scalar_get = butlerQC.get(refs[0])
                self.assertEqual(scalar_get, list_get[0])

                with self.assertRaises(TypeError):
                    butlerQC.get({})

                # Output ref won't be known to this quantum.
                outputs = [ref for _, ref in outputRefs]
                with self.assertRaises(ValueError):
                    butlerQC.get(outputs[0])

        # look at the output produced by the task
        dsdata = butler.datasets[output_name]
        self.assertEqual(len(dsdata), len(quanta))
        for i, quantum in enumerate(quanta):
            ref = quantum.outputs[output_name][0]
            self.assertEqual(dsdata[ref.dataId], 100 + i + 3)

    def testChain2Full(self) -> None:
        """Test for two-task chain with full butler."""
        self._testChain2(full_butler=True)

    def testChain2Limited(self) -> None:
        """Test for two-task chain with limited butler."""
        self._testChain2(full_butler=False)

    def _testChain2(self, full_butler: bool) -> None:
        """Test for two-task chain."""
        butler = ButlerMock()
        config1 = AddConfig()
        overall_input_name = config1.connections.input
        intermediate_name = config1.connections.output
        task1 = AddTask(config=config1)
        config2 = AddConfig()
        config2.addend = 200
        config2.connections.input = intermediate_name
        overall_output_name = "add_output_2"
        config2.connections.output = overall_output_name
        task2 = AddTask2(config=config2)

        pipeline_graph = pipeBase.PipelineGraph()
        task1_node = pipeline_graph.add_task(None, type(task1), task1.config)
        task2_node = pipeline_graph.add_task(None, type(task2), task2.config)
        pipeline_graph.resolve(dimensions=butler.dimensions, dataset_types={})

        # make all quanta
        quanta1 = self._makeQuanta(task1_node, pipeline_graph)
        quanta2 = self._makeQuanta(task2_node, pipeline_graph)

        # add input data to butler
        dstype0 = pipeline_graph.dataset_types[overall_input_name].dataset_type
        for i, quantum in enumerate(quanta1):
            ref = quantum.inputs[dstype0.name][0]
            butler.put(100 + i, ref)

        # run task on each quanta
        for quantum in quanta1:
            butlerQC = pipeBase.QuantumContext(butler, quantum)
            inputRefs, outputRefs = task1_node.get_connections().buildDatasetRefs(quantum)
            task1.runQuantum(butlerQC, inputRefs, outputRefs)
        for quantum in quanta2:
            butlerQC = pipeBase.QuantumContext(butler, quantum)
            inputRefs, outputRefs = task2_node.get_connections().buildDatasetRefs(quantum)
            task2.runQuantum(butlerQC, inputRefs, outputRefs)

        # look at the output produced by the task
        dsdata = butler.datasets[intermediate_name]
        self.assertEqual(len(dsdata), len(quanta1))
        for i, quantum in enumerate(quanta1):
            ref = quantum.outputs[intermediate_name][0]
            self.assertEqual(dsdata[ref.dataId], 100 + i + 3)

        dsdata = butler.datasets[overall_output_name]
        self.assertEqual(len(dsdata), len(quanta2))
        for i, quantum in enumerate(quanta2):
            ref = quantum.outputs[overall_output_name][0]
            self.assertEqual(dsdata[ref.dataId], 100 + i + 3 + 200)

    def testButlerQC(self):
        """Test for QuantumContext. Full and limited share get
        implementation so only full is tested.
        """
        butler = ButlerMock()
        task = AddTask(config=AddConfig())
        input_name = task.config.connections.input

        pipeline_graph = pipeBase.PipelineGraph()
        task_node = pipeline_graph.add_task(None, type(task), config=task.config)
        pipeline_graph.resolve(dimensions=butler.dimensions, dataset_types={})

        # make one quantum
        (quantum,) = self._makeQuanta(task_node, pipeline_graph, 1)

        # add input data to butler
        dstype0 = pipeline_graph.dataset_types[input_name].dataset_type
        ref = quantum.inputs[dstype0.name][0]
        butler.put(100, ref)

        butlerQC = pipeBase.QuantumContext(butler, quantum)
        self.assertEqual(butlerQC.resources.num_cores, 1)
        self.assertIsNone(butlerQC.resources.max_mem)

        # Pass ref as single argument or a list.
        obj = butlerQC.get(ref)
        self.assertEqual(obj, 100)
        obj = butlerQC.get([ref])
        self.assertEqual(obj, [100])

        # Pass None instead of a ref.
        obj = butlerQC.get(None)
        self.assertIsNone(obj)
        obj = butlerQC.get([None])
        self.assertEqual(obj, [None])

        # COmbine a ref and None.
        obj = butlerQC.get([ref, None])
        self.assertEqual(obj, [100, None])

        # Use refs from a QuantizedConnection.
        inputRefs, outputRefs = task_node.get_connections().buildDatasetRefs(quantum)
        obj = butlerQC.get(inputRefs)
        self.assertEqual(obj, {"input": 100})

        # Add few None values to a QuantizedConnection.
        inputRefs.input = [None, ref]
        inputRefs.input2 = None
        obj = butlerQC.get(inputRefs)
        self.assertEqual(obj, {"input": [None, 100], "input2": None})

        # Set additional context.
        resources = pipeBase.ExecutionResources(num_cores=4, max_mem=5 * u.MB)
        butlerQC = pipeBase.QuantumContext(butler, quantum, resources=resources)
        self.assertEqual(butlerQC.resources.num_cores, 4)
        self.assertEqual(butlerQC.resources.max_mem, 5_000_000 * u.B)

        resources = pipeBase.ExecutionResources(max_mem=5)
        butlerQC = pipeBase.QuantumContext(butler, quantum, resources=resources)
        self.assertEqual(butlerQC.resources.num_cores, 1)
        self.assertEqual(butlerQC.resources.max_mem, 5 * u.B)

    def test_ExecutionResources(self):
        res = pipeBase.ExecutionResources()
        self.assertEqual(res.num_cores, 1)
        self.assertIsNone(res.max_mem)
        self.assertEqual(pickle.loads(pickle.dumps(res)), res)

        res = pipeBase.ExecutionResources(num_cores=4, max_mem=1 * u.MiB)
        self.assertEqual(res.num_cores, 4)
        self.assertEqual(res.max_mem.value, 1024 * 1024)
        self.assertEqual(pickle.loads(pickle.dumps(res)), res)

        res = pipeBase.ExecutionResources(max_mem=512)
        self.assertEqual(res.num_cores, 1)
        self.assertEqual(res.max_mem.value, 512)
        self.assertEqual(pickle.loads(pickle.dumps(res)), res)

        res = pipeBase.ExecutionResources(max_mem="")
        self.assertIsNone(res.max_mem)

        res = pipeBase.ExecutionResources(max_mem="32 KiB")
        self.assertEqual(res.num_cores, 1)
        self.assertEqual(res.max_mem.value, 32 * 1024)
        self.assertEqual(pickle.loads(pickle.dumps(res)), res)

        self.assertIs(res, copy.deepcopy(res))

        with self.assertRaises(AttributeError):
            res.num_cores = 4

        with self.assertRaises(u.UnitConversionError):
            pipeBase.ExecutionResources(max_mem=1 * u.m)
        with self.assertRaises(ValueError):
            pipeBase.ExecutionResources(num_cores=-32)
        with self.assertRaises(u.UnitConversionError):
            pipeBase.ExecutionResources(max_mem=1 * u.m)
        with self.assertRaises(u.UnitConversionError):
            pipeBase.ExecutionResources(max_mem=1, default_mem_units=u.m)
        with self.assertRaises(u.UnitConversionError):
            pipeBase.ExecutionResources(max_mem="32 Pa")


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Run file leak tests."""


def setup_module(module):
    """Configure pytest."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import abc
import logging
from typing import Protocol

from lsst.utils import introspection

from ._task_metadata import GetSetDictMetadata, NestedMetadataDict

__all__ = (
    "UnprocessableDataError",
    "AnnotatedPartialOutputsError",
    "NoWorkFound",
    "RepeatableQuantumError",
    "AlgorithmError",
    "InvalidQuantumError",
)


class GetSetDictMetadataHolder(Protocol):
    """Protocol for objects that have a ``metadata`` attribute that satisfies
    `GetSetDictMetadata`.
    """

    metadata: GetSetDictMetadata | None


class NoWorkFound(BaseException):
    """An exception raised when a Quantum should not exist because there is no
    work for it to do.

    This usually occurs because a non-optional input dataset is not present, or
    a spatiotemporal overlap that was conservatively predicted does not
    actually exist.

    This inherits from BaseException because it is used to signal a case that
    we don't consider a real error, even though we often want to use try/except
    logic to trap it.
    """


class RepeatableQuantumError(RuntimeError):
    """Exception that may be raised by PipelineTasks (and code they delegate
    to) in order to indicate that a repeatable problem that will not be
    addressed by retries.

    This usually indicates that the algorithm and the data it has been given
    are somehow incompatible, and the task should run fine on most other data.

    This exception may be used as a base class for more specific questions, or
    used directly while chaining another exception, e.g.::

        try:
            run_code()
        except SomeOtherError as err:
            raise RepeatableQuantumError() from err

    This may be used for missing input data when the desired behavior is to
    cause all downstream tasks being run be blocked, forcing the user to
    address the problem.  When the desired behavior is to skip all of this
    quantum and attempt downstream tasks (or skip them) without its its
    outputs, raise `NoWorkFound` or return without raising instead.
    """

    EXIT_CODE = 20


class AlgorithmError(RepeatableQuantumError, abc.ABC):
    """Exception that may be raised by PipelineTasks (and code they delegate
    to) in order to indicate a repeatable algorithmic failure that will not be
    addressed by retries.

    Subclass this exception to define the metadata associated with the error
    (for example: number of data points in a fit vs. degrees of freedom).
    """

    @property
    @abc.abstractmethod
    def metadata(self) -> NestedMetadataDict | None:
        """Metadata from the raising `~lsst.pipe.base.Task` with more
        information about the failure. The contents of the dict are
        `~lsst.pipe.base.Task`-dependent, and must have `str` keys and `str`,
        `int`, `float`, `bool`, or nested-dictionary (with the same key and
        value types) values.
        """
        raise NotImplementedError


class UnprocessableDataError(RepeatableQuantumError):
    """Exception that will be subclassed and raised by Tasks to indicate a
    failure to process their inputs for some reason that is non-recoverable.

    Notes
    -----
    An example is a known bright star that causes PSF measurement to fail, and
    that makes that detector entirely non-recoverable.

    Do not raise this unless we are convinced that the data cannot be
    processed, even by a better algorithm. Most instances where this error
    would be raised likely require an RFC to explicitly define the situation.
    """


class AnnotatedPartialOutputsError(RepeatableQuantumError):
    """Exception that runQuantum raises when the (partial) outputs it has
    written contain information about their own incompleteness or degraded
    quality.

    This exception should always chain the original error. When the
    executor catches this exception, it will report the original exception. In
    contrast, other exceptions raised from ``runQuantum`` are considered to
    invalidate any outputs that are already written.
    """

    @classmethod
    def annotate(
        cls, error: Exception, *args: GetSetDictMetadataHolder | None, log: logging.Logger
    ) -> AnnotatedPartialOutputsError:
        """Set metadata on outputs to explain the nature of the failure.

        Parameters
        ----------
        error : `Exception`
            Exception that caused the task to fail.
        *args : `GetSetDictMetadataHolder`
            Objects (e.g. Task, Exposure, SimpleCatalog) to annotate with
            failure information. They must have a `metadata` property.
        log : `logging.Logger`
            Log to send error message to.

        Returns
        -------
        error : `AnnotatedPartialOutputsError`
            Exception that the failing task can ``raise from`` with the
            passed-in exception.

        Notes
        -----
        This should be called from within an except block that has caught an
        exception. Here is an example of handling a failure in
        ``PipelineTask.runQuantum`` that annotates and writes partial outputs:

        .. code-block:: py
            :name: annotate-error-example

            def runQuantum(self, butlerQC, inputRefs, outputRefs):
                inputs = butlerQC.get(inputRefs)
                exposures = inputs.pop("exposures")
                assert not inputs, "runQuantum got more inputs than expected"

                result = pipeBase.Struct(catalog=None)
                try:
                    self.run(exposure)
                except pipeBase.AlgorithmError as e:
                    error = pipeBase.AnnotatedPartialOutputsError.annotate(
                        e, self, result.catalog, log=self.log
                    )
                    raise error from e
                finally:
                    butlerQC.put(result, outputRefs)
        """
        failure_info = {
            "message": str(error),
            "type": introspection.get_full_type_name(error),
        }
        if other := getattr(error, "metadata", None):
            failure_info["metadata"] = other

        # NOTE: Can't fully test this in pipe_base because afw is not a
        # dependency; test_calibrateImage.py in pipe_tasks gives more coverage.
        for item in args:
            # Some outputs may not exist, so we cannot set metadata on them.
            if item is None:
                continue
            item.metadata.set_dict("failure", failure_info)  # type: ignore

        log.exception(
            "Task failed with only partial outputs; see exception message for details.",
            exc_info=error,
        )

        return cls("Task failed and wrote partial outputs: see chained exception for details.")


class InvalidQuantumError(Exception):
    """Exception that may be raised by PipelineTasks (and code they delegate
    to) in order to indicate logic bug or configuration problem.

    This usually indicates that the configured algorithm itself is invalid and
    will not run on a significant fraction of quanta (often all of them).

    This exception may be used as a base class for more specific questions, or
    used directly while chaining another exception, e.g.::

        try:
            run_code()
        except SomeOtherError as err:
            raise RepeatableQuantumError() from err

    Raising this exception in `PipelineTask.runQuantum` or something it calls
    is a last resort - whenever possible, such problems should cause exceptions
    in ``__init__`` or in QuantumGraph generation.  It should never be used
    for missing data.
    """

    EXIT_CODE = 21



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for Pipeline.
"""

import pickle
import textwrap
import unittest

import lsst.utils.tests
from lsst.pipe.base import Pipeline, TaskDef
from lsst.pipe.base.pipelineIR import LabeledSubset
from lsst.pipe.base.tests.simpleQGraph import AddTask, makeSimplePipeline


class PipelineTestCase(unittest.TestCase):
    """A test case for TaskDef and Pipeline."""

    def testTaskDef(self):
        """Tests for TaskDef structure."""
        task1 = TaskDef(taskClass=AddTask, config=AddTask.ConfigClass())
        self.assertIn("Add", task1.taskName)
        self.assertIsInstance(task1.config, AddTask.ConfigClass)
        self.assertIsNotNone(task1.taskClass)
        self.assertEqual(task1.label, "add_task")
        task1a = pickle.loads(pickle.dumps(task1))
        self.assertEqual(task1, task1a)

    def testEmpty(self):
        """Creating empty pipeline."""
        pipeline = Pipeline("test")
        self.assertEqual(len(pipeline), 0)

    def testInitial(self):
        """Testing constructor with initial data."""
        pipeline = makeSimplePipeline(2)
        self.assertEqual(len(pipeline), 2)
        pipeline_graph = pipeline.to_graph()
        pipeline_graph.sort()
        task_nodes = list(pipeline_graph.tasks.values())
        self.assertEqual(task_nodes[0].task_class, AddTask)
        self.assertEqual(task_nodes[1].task_class, AddTask)
        self.assertEqual(task_nodes[0].label, "task0")
        self.assertEqual(task_nodes[1].label, "task1")
        self.assertEqual(pipeline.task_labels, {"task0", "task1"})

    def testModifySubset(self):
        pipeline = makeSimplePipeline(2)

        # Test adding labels.
        with self.assertRaises(ValueError):
            pipeline.addLabelToSubset("test", "new_label")
        pipeline._pipelineIR.labeled_subsets["test"] = LabeledSubset("test", set(), None)
        with self.assertRaises(ValueError):
            pipeline.addLabelToSubset("test", "missing_label")
        pipeline.addLabelToSubset("test", "task0")
        self.assertEqual(pipeline._pipelineIR.labeled_subsets["test"].subset, {"task0"})

        # Test removing labels.
        with self.assertRaises(ValueError):
            pipeline.addLabelToSubset("missing_subset", "task0")
        with self.assertRaises(ValueError):
            pipeline.addLabelToSubset("test", "missing_label")
        pipeline.removeLabelFromSubset("test", "task0")
        self.assertEqual(pipeline._pipelineIR.labeled_subsets["test"].subset, set())

        # Test creating new labeled subsets.
        with self.assertRaises(ValueError):
            # missing task label
            pipeline.addLabeledSubset("newSubset", "test description", {"missing_task_label"})
        with self.assertRaises(ValueError):
            # duplicate labeled subset
            pipeline.addLabeledSubset("test", "test description", {"missing_task_label"})

        taskLabels = {"task0", "task1"}
        pipeline.addLabeledSubset("newSubset", "test description", taskLabels)

        # verify using the subset property interface
        self.assertEqual(pipeline.subsets["newSubset"], taskLabels)

        # Test removing labeled subsets
        with self.assertRaises(ValueError):
            pipeline.removeLabeledSubset("missing_subset")

        pipeline.removeLabeledSubset("newSubset")
        self.assertNotIn("newSubset", pipeline.subsets.keys())

    def testMergingPipelines(self):
        pipeline1 = makeSimplePipeline(2)
        pipeline2 = makeSimplePipeline(4)
        pipeline2.removeTask("task0")
        pipeline2.removeTask("task1")

        pipeline1.mergePipeline(pipeline2)
        self.assertEqual(pipeline1._pipelineIR.tasks.keys(), {"task0", "task1", "task2", "task3"})

    def testFindingSubset(self):
        pipeline = makeSimplePipeline(2)
        pipeline._pipelineIR.labeled_subsets["test1"] = LabeledSubset("test1", set(), None)
        pipeline._pipelineIR.labeled_subsets["test2"] = LabeledSubset("test2", set(), None)
        pipeline._pipelineIR.labeled_subsets["test3"] = LabeledSubset("test3", set(), None)

        pipeline.addLabelToSubset("test1", "task0")
        pipeline.addLabelToSubset("test3", "task0")

        with self.assertRaises(ValueError):
            pipeline.findSubsetsWithLabel("missing_label")

        self.assertEqual(pipeline.findSubsetsWithLabel("task0"), {"test1", "test3"})

    def testParameters(self):
        """Test that parameters can be set and used to format."""
        pipeline_str = textwrap.dedent(
            """
            description: Test Pipeline
            parameters:
               testValue: 5
            tasks:
              add:
                class: test_pipeline.AddTask
                config:
                  addend: parameters.testValue
        """
        )
        # verify that parameters are used in expanding a pipeline
        pipeline = Pipeline.fromString(pipeline_str)
        pipeline_graph = pipeline.to_graph()
        self.assertEqual(pipeline_graph.tasks["add"].config.addend, 5)

        # verify that a parameter can be overridden on the "command line"
        pipeline.addConfigOverride("parameters", "testValue", 14)
        pipeline_graph = pipeline.to_graph()
        self.assertEqual(pipeline_graph.tasks["add"].config.addend, 14)

        # verify that parameters does not support files or python overrides
        with self.assertRaises(ValueError):
            pipeline.addConfigFile("parameters", "fakeFile")
        with self.assertRaises(ValueError):
            pipeline.addConfigPython("parameters", "fakePythonString")

    def testSerialization(self):
        pipeline = makeSimplePipeline(2)
        dump = str(pipeline)
        load = Pipeline.fromString(dump)
        self.assertEqual(pipeline, load)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Run file leak tests."""


def setup_module(module):
    """Configure pytest."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining TaskFactory interface.
"""

from __future__ import annotations

__all__ = ["TaskFactory"]

from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lsst.daf.butler import DatasetRef, LimitedButler

    from .pipeline import TaskDef
    from .pipeline_graph import TaskNode
    from .pipelineTask import PipelineTask


class TaskFactory(metaclass=ABCMeta):
    """Abstract base class for task factory.

    Task factory is responsible for creating instances of PipelineTask
    subclasses.
    """

    @abstractmethod
    def makeTask(
        self,
        task_node: TaskDef | TaskNode,  # TODO: remove TaskDef on DM-40443.
        /,
        butler: LimitedButler,
        initInputRefs: Iterable[DatasetRef] | None,
    ) -> PipelineTask:
        """Create new PipelineTask instance from its `~lsst.pipe.base.TaskDef`.

        Parameters
        ----------
        task_node : `~pipeline_graph.TaskNode` or `TaskDef`
            Task definition structure.  `TaskDef` support is deprecated and
            will be removed after v27.
        butler : `lsst.daf.butler.LimitedButler`
            Butler instance used to obtain initialization inputs for task.
        initInputRefs : `~collections.abc.Iterable` of \
                `~lsst.daf.butler.DatasetRef` or `None`
            List of resolved dataset references for init inputs for this task.

        Returns
        -------
        task : `PipelineTask`
            Instance of a PipelineTask class.

        Raises
        ------
        Any exceptions that are raised by PipelineTask constructor or its
        configuration class are propagated back to caller.
        """
        raise NotImplementedError()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("DESERIALIZER_MAP",)

import json
import lzma
import pickle
import struct
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from types import SimpleNamespace
from typing import TYPE_CHECKING, ClassVar, cast

import networkx as nx
from lsst.daf.butler import (
    DatasetRef,
    DatasetType,
    DimensionConfig,
    DimensionUniverse,
    Quantum,
    SerializedDimensionRecord,
)
from lsst.utils import doImportType

from ..config import PipelineTaskConfig
from ..pipeline import TaskDef
from ..pipelineTask import PipelineTask
from ._implDetails import DatasetTypeName, _DatasetTracker
from .quantumNode import QuantumNode, SerializedQuantumNode

if TYPE_CHECKING:
    from .graph import QuantumGraph


class StructSizeDescriptor:
    """Class level property. It exists to report the size
    (number of bytes) of whatever the formatter string is for a deserializer.
    """

    def __get__(self, inst: DeserializerBase | None, owner: type[DeserializerBase]) -> int:
        return struct.calcsize(owner.FMT_STRING())


@dataclass
class DeserializerBase(ABC):
    @classmethod
    @abstractmethod
    def FMT_STRING(cls) -> str:
        raise NotImplementedError("Base class does not implement this method")

    structSize: ClassVar[StructSizeDescriptor]

    preambleSize: int
    sizeBytes: bytes

    def __init_subclass__(cls) -> None:
        # attach the size decriptor
        cls.structSize = StructSizeDescriptor()
        super().__init_subclass__()

    def unpackHeader(self, rawHeader: bytes) -> str | None:
        """Transform the raw bytes corresponding to the header of a save into
        a string of the header information.

        Parameters
        ----------
        rawHeader : bytes
            The bytes that are to be parsed into the header information. These
            are the bytes after the preamble and structsize number of bytes
            and before the headerSize bytes.

        Returns
        -------
        header : `str` or `None`
            Header information as a string. Returns `None` if the save format
            has no header string implementation (such as save format 1 that is
            all pickle).
        """
        raise NotImplementedError("Base class does not implement this method")

    @property
    def headerSize(self) -> int:
        """Returns the number of bytes from the beginning of the file to the
        end of the metadata.
        """
        raise NotImplementedError("Base class does not implement this method")

    def readHeaderInfo(self, rawHeader: bytes) -> SimpleNamespace:
        """Parse the supplied raw bytes into the header information and
        byte ranges of specific TaskDefs and QuantumNodes.

        Parameters
        ----------
        rawHeader : bytes
            The bytes that are to be parsed into the header information. These
            are the bytes after the preamble and structsize number of bytes
            and before the headerSize bytes.
        """
        raise NotImplementedError("Base class does not implement this method")

    def constructGraph(
        self,
        nodes: set[uuid.UUID],
        _readBytes: Callable[[int, int], bytes],
        universe: DimensionUniverse | None = None,
    ) -> QuantumGraph:
        """Construct a graph from the deserialized information.

        Parameters
        ----------
        nodes : `set` of `uuid.UUID`
            The nodes to include in the graph.
        _readBytes : callable
            A callable that can be used to read bytes from the file handle.
            The callable will take two ints, start and stop, to use as the
            numerical bounds to read and returns a byte stream.
        universe : `~lsst.daf.butler.DimensionUniverse`
            The singleton of all dimensions known to the middleware registry.
        """
        raise NotImplementedError("Base class does not implement this method")

    def description(self) -> str:
        """Return the description of the serialized data format.

        Returns
        -------
        desc : `str`
            Description of serialized data format.
        """
        raise NotImplementedError("Base class does not implement this method")


Version1Description = """
The save file starts with the first few bytes corresponding to the magic bytes
in the QuantumGraph: `qgraph4\xf6\xe8\xa9`.

The next few bytes are 2 big endian unsigned 64 bit integers.

The first unsigned 64 bit integer corresponds to the number of bytes of a
python mapping of TaskDef labels to the byte ranges in the save file where the
definition can be loaded.

The second unsigned 64 bit integer corrresponds to the number of bytes of a
python mapping of QuantumGraph Node number to the byte ranges in the save file
where the node can be loaded. The byte range is indexed starting after
the `header` bytes of the magic bytes, size bytes, and bytes of the two
mappings.

Each of the above mappings are pickled and then lzma compressed, so to
deserialize the bytes, first lzma decompression must be performed and the
results passed to python pickle loader.

As stated above, each map contains byte ranges of the corresponding
datastructure. Theses bytes are also lzma compressed pickles, and should
be deserialized in a similar manner. The byte range is indexed starting after
the `header` bytes of the magic bytes, size bytes, and bytes of the two
mappings.

In addition to the the TaskDef byte locations, the TypeDef map also contains
an additional key '__GraphBuildID'. The value associated with this is the
unique id assigned to the graph at its creation time.
"""


@dataclass
class DeserializerV1(DeserializerBase):
    @classmethod
    def FMT_STRING(cls) -> str:
        return ">QQ"

    def __post_init__(self) -> None:
        self.taskDefMapSize, self.nodeMapSize = struct.unpack(self.FMT_STRING(), self.sizeBytes)

    @property
    def headerSize(self) -> int:
        return self.preambleSize + self.structSize + self.taskDefMapSize + self.nodeMapSize

    def readHeaderInfo(self, rawHeader: bytes) -> SimpleNamespace:
        returnValue = SimpleNamespace()
        returnValue.taskDefMap = pickle.loads(rawHeader[: self.taskDefMapSize])
        returnValue._buildId = returnValue.taskDefMap["__GraphBuildID"]
        returnValue.map = pickle.loads(rawHeader[self.taskDefMapSize :])
        returnValue.metadata = None
        self.returnValue = returnValue
        return returnValue

    def unpackHeader(self, rawHeader: bytes) -> str | None:
        return None

    def constructGraph(
        self,
        nodes: set[uuid.UUID],
        _readBytes: Callable[[int, int], bytes],
        universe: DimensionUniverse | None = None,
    ) -> QuantumGraph:
        # need to import here to avoid cyclic imports
        from . import QuantumGraph

        quanta: defaultdict[TaskDef, set[Quantum]] = defaultdict(set)
        quantumToNodeId: dict[Quantum, uuid.UUID] = {}
        loadedTaskDef = {}
        # loop over the nodes specified above
        for node in nodes:
            # Get the bytes to read from the map
            start, stop = self.returnValue.map[node]
            start += self.headerSize
            stop += self.headerSize

            # read the specified bytes, will be overloaded by subclasses
            # bytes are compressed, so decompress them
            dump = lzma.decompress(_readBytes(start, stop))

            # reconstruct node
            qNode = pickle.loads(dump)
            object.__setattr__(qNode, "nodeId", uuid.uuid4())

            # read the saved node, name. If it has been loaded, attach it, if
            # not read in the taskDef first, and then load it
            nodeTask = qNode.taskDef
            if nodeTask not in loadedTaskDef:
                # Get the byte ranges corresponding to this taskDef
                start, stop = self.returnValue.taskDefMap[nodeTask]
                start += self.headerSize
                stop += self.headerSize

                # load the taskDef, this method call will be overloaded by
                # subclasses.
                # bytes are compressed, so decompress them
                taskDef = pickle.loads(lzma.decompress(_readBytes(start, stop)))
                loadedTaskDef[nodeTask] = taskDef
            # Explicitly overload the "frozen-ness" of nodes to attach the
            # taskDef back into the un-persisted node
            object.__setattr__(qNode, "taskDef", loadedTaskDef[nodeTask])
            quanta[qNode.taskDef].add(qNode.quantum)

            # record the node for later processing
            quantumToNodeId[qNode.quantum] = qNode.nodeId

        # construct an empty new QuantumGraph object, and run the associated
        # creation method with the un-persisted data
        qGraph = object.__new__(QuantumGraph)
        qGraph._buildGraphs(
            quanta,
            _quantumToNodeId=quantumToNodeId,
            _buildId=self.returnValue._buildId,
            metadata=self.returnValue.metadata,
            universe=universe,
        )
        return qGraph

    def description(self) -> str:
        return Version1Description


Version2Description = """
The save file starts with the first few bytes corresponding to the magic bytes
in the QuantumGraph: `qgraph4\xf6\xe8\xa9`.

The next few bytes are a big endian unsigned long long.

The unsigned long long corresponds to the number of bytes of a python mapping
of header information. This mapping is encoded into json and then lzma
compressed, meaning the operations must be performed in the opposite order to
deserialize.

The json encoded header mapping contains 4 fields: TaskDefs, GraphBuildId,
Nodes, and Metadata.

The `TaskDefs` key corresponds to a value which is a mapping of Task label to
task data. The task data is a mapping of key to value, where the only key is
`bytes` and it corresponds to a tuple of a byte range of the start, stop
bytes (indexed after all the header bytes)

The `GraphBuildId` corresponds with a string that is the unique id assigned to
this graph when it was created.

The `Nodes` key is like the `TaskDefs` key except it corresponds to
QuantumNodes instead of TaskDefs. Another important difference is that JSON
formatting does not allow using numbers as keys, and this mapping is keyed by
the node number. Thus it is stored in JSON as two equal length lists, the first
being the keys, and the second the values associated with those keys.

The `Metadata` key is a mapping of strings to associated values. This metadata
may be anything that is important to be transported alongside the graph.

As stated above, each map contains byte ranges of the corresponding
datastructure. Theses bytes are also lzma compressed pickles, and should
be deserialized in a similar manner.
"""


@dataclass
class DeserializerV2(DeserializerBase):
    @classmethod
    def FMT_STRING(cls) -> str:
        return ">Q"

    def __post_init__(self) -> None:
        (self.mapSize,) = struct.unpack(self.FMT_STRING(), self.sizeBytes)

    @property
    def headerSize(self) -> int:
        return self.preambleSize + self.structSize + self.mapSize

    def readHeaderInfo(self, rawHeader: bytes) -> SimpleNamespace:
        uncompressedHeaderMap = self.unpackHeader(rawHeader)
        if uncompressedHeaderMap is None:
            raise ValueError(
                "This error is not possible because self.unpackHeader cannot return None,"
                " but is done to satisfy type checkers"
            )
        header = json.loads(uncompressedHeaderMap)
        returnValue = SimpleNamespace()
        returnValue.taskDefMap = header["TaskDefs"]
        returnValue._buildId = header["GraphBuildID"]
        returnValue.map = dict(header["Nodes"])
        returnValue.metadata = header["Metadata"]
        self.returnValue = returnValue
        return returnValue

    def unpackHeader(self, rawHeader: bytes) -> str | None:
        return lzma.decompress(rawHeader).decode()

    def constructGraph(
        self,
        nodes: set[uuid.UUID],
        _readBytes: Callable[[int, int], bytes],
        universe: DimensionUniverse | None = None,
    ) -> QuantumGraph:
        # need to import here to avoid cyclic imports
        from . import QuantumGraph

        quanta: defaultdict[TaskDef, set[Quantum]] = defaultdict(set)
        quantumToNodeId: dict[Quantum, uuid.UUID] = {}
        loadedTaskDef = {}
        # loop over the nodes specified above
        for node in nodes:
            # Get the bytes to read from the map
            start, stop = self.returnValue.map[node]["bytes"]
            start += self.headerSize
            stop += self.headerSize

            # read the specified bytes, will be overloaded by subclasses
            # bytes are compressed, so decompress them
            dump = lzma.decompress(_readBytes(start, stop))

            # reconstruct node
            qNode = pickle.loads(dump)
            object.__setattr__(qNode, "nodeId", uuid.uuid4())

            # read the saved node, name. If it has been loaded, attach it, if
            # not read in the taskDef first, and then load it
            nodeTask = qNode.taskDef
            if nodeTask not in loadedTaskDef:
                # Get the byte ranges corresponding to this taskDef
                start, stop = self.returnValue.taskDefMap[nodeTask]["bytes"]
                start += self.headerSize
                stop += self.headerSize

                # load the taskDef, this method call will be overloaded by
                # subclasses.
                # bytes are compressed, so decompress them
                taskDef = pickle.loads(lzma.decompress(_readBytes(start, stop)))
                loadedTaskDef[nodeTask] = taskDef
            # Explicitly overload the "frozen-ness" of nodes to attach the
            # taskDef back into the un-persisted node
            object.__setattr__(qNode, "taskDef", loadedTaskDef[nodeTask])
            quanta[qNode.taskDef].add(qNode.quantum)

            # record the node for later processing
            quantumToNodeId[qNode.quantum] = qNode.nodeId

        # construct an empty new QuantumGraph object, and run the associated
        # creation method with the un-persisted data
        qGraph = object.__new__(QuantumGraph)
        qGraph._buildGraphs(
            quanta,
            _quantumToNodeId=quantumToNodeId,
            _buildId=self.returnValue._buildId,
            metadata=self.returnValue.metadata,
            universe=universe,
        )
        return qGraph

    def description(self) -> str:
        return Version2Description


Version3Description = """
The save file starts with the first few bytes corresponding to the magic bytes
in the QuantumGraph: `qgraph4\xf6\xe8\xa9`.

The next few bytes are a big endian unsigned long long.

The unsigned long long corresponds to the number of bytes of a mapping
of header information. This mapping is encoded into json and then lzma
compressed, meaning the operations must be performed in the opposite order to
deserialize.

The json encoded header mapping contains 5 fields: GraphBuildId, TaskDefs,
Nodes, Metadata, and DimensionRecords.

The `GraphBuildId` key corresponds with a string that is the unique id assigned
to this graph when it was created.

The `TaskDefs` key corresponds to a value which is a mapping of Task label to
task data. The task data is a mapping of key to value. The keys of this mapping
are `bytes`, `inputs`, and `outputs`.

The  `TaskDefs` `bytes` key corresponds to a tuple of a byte range of the
start, stop bytes (indexed after all the header bytes).  This byte rage
corresponds to a lzma compressed json mapping. This mapping has keys of
`taskName`, corresponding to a fully qualified python class, `config` a
pex_config string that is used to configure the class, and `label` which
corresponds to a string that uniquely identifies the task within a given
execution pipeline.

The `TaskDefs` `inputs` key is associated with a list of tuples where each
tuple is a label of a task that is considered coming before a given task, and
the name of the dataset that is shared between the tasks (think node and edge
in a graph sense).

The `TaskDefs` `outputs` key is like inputs except the values in a list
correspond to all the output connections of a task.

The `Nodes` key is also a json mapping with keys corresponding to the UUIDs of
QuantumNodes. The values associated with these keys is another mapping with
the keys `bytes`, `inputs`, and `outputs`.

`Nodes` key `bytes` corresponds to a tuple of a byte range of the start, stop
bytes (indexed after all the header bytes). These bytes are a lzma compressed
json mapping which contains many sub elements, this mapping will be referred to
as the SerializedQuantumNode (related to the python class it corresponds to).

SerializedQUantumNodes have 3 keys, `quantum` corresponding to a json mapping
(described below) referred to as a SerializedQuantum, `taskLabel` a string
which corresponds to a label in the `TaskDefs` mapping, and `nodeId.

A SerializedQuantum has many keys; taskName, dataId, datasetTypeMapping,
initInputs, inputs, outputs, dimensionRecords.

like the `TaskDefs` key except it corresponds to
QuantumNodes instead of TaskDefs, and the keys of the mappings are string
representations of the UUIDs of the QuantumNodes.

The `Metadata` key is a mapping of strings to associated values. This metadata
may be anything that is important to be transported alongside the graph.

As stated above, each map contains byte ranges of the corresponding
datastructure. Theses bytes are also lzma compressed pickles, and should
be deserialized in a similar manner.
"""


@dataclass
class DeserializerV3(DeserializerBase):
    @classmethod
    def FMT_STRING(cls) -> str:
        return ">Q"

    def __post_init__(self) -> None:
        self.infoSize: int
        (self.infoSize,) = struct.unpack(self.FMT_STRING(), self.sizeBytes)

    @property
    def headerSize(self) -> int:
        return self.preambleSize + self.structSize + self.infoSize

    def readHeaderInfo(self, rawHeader: bytes) -> SimpleNamespace:
        uncompressedinfoMap = self.unpackHeader(rawHeader)
        assert uncompressedinfoMap is not None  # for python typing, this variant can't be None
        infoMap = json.loads(uncompressedinfoMap)
        infoMappings = SimpleNamespace()
        infoMappings.taskDefMap = infoMap["TaskDefs"]
        infoMappings._buildId = infoMap["GraphBuildID"]
        infoMappings.map = {uuid.UUID(k): v for k, v in infoMap["Nodes"]}
        infoMappings.metadata = infoMap["Metadata"]
        infoMappings.dimensionRecords = {}
        for k, v in infoMap["DimensionRecords"].items():
            infoMappings.dimensionRecords[int(k)] = SerializedDimensionRecord(**v)
        # This is important to be a get call here, so that it supports versions
        # of saved quantum graph that might not have a saved universe without
        # changing save format
        if (universeConfig := infoMap.get("universe")) is not None:
            universe = DimensionUniverse(config=DimensionConfig(universeConfig))
        else:
            universe = DimensionUniverse()
        infoMappings.universe = universe
        infoMappings.globalInitOutputRefs = []
        if (json_refs := infoMap.get("GlobalInitOutputRefs")) is not None:
            infoMappings.globalInitOutputRefs = [
                DatasetRef.from_json(json_ref, universe=universe) for json_ref in json_refs
            ]
        infoMappings.registryDatasetTypes = []
        if (json_refs := infoMap.get("RegistryDatasetTypes")) is not None:
            infoMappings.registryDatasetTypes = [
                DatasetType.from_json(json_ref, universe=universe) for json_ref in json_refs
            ]
        self.infoMappings = infoMappings
        return infoMappings

    def unpackHeader(self, rawHeader: bytes) -> str | None:
        return lzma.decompress(rawHeader).decode()

    def constructGraph(
        self,
        nodes: set[uuid.UUID],
        _readBytes: Callable[[int, int], bytes],
        universe: DimensionUniverse | None = None,
    ) -> QuantumGraph:
        # need to import here to avoid cyclic imports
        from . import QuantumGraph

        graph = nx.DiGraph()
        loadedTaskDef: dict[str, TaskDef] = {}
        container = {}
        datasetDict = _DatasetTracker(createInverse=True)
        taskToQuantumNode: defaultdict[TaskDef, set[QuantumNode]] = defaultdict(set)
        initInputRefs: dict[TaskDef, list[DatasetRef]] = {}
        initOutputRefs: dict[TaskDef, list[DatasetRef]] = {}

        if universe is not None:
            if not universe.isCompatibleWith(self.infoMappings.universe):
                saved = self.infoMappings.universe
                raise RuntimeError(
                    f"The saved dimension universe ({saved.namespace}@v{saved.version}) is not "
                    f"compatible with the supplied universe ({universe.namespace}@v{universe.version})."
                )
        else:
            universe = self.infoMappings.universe

        for node in nodes:
            start, stop = self.infoMappings.map[node]["bytes"]
            start, stop = start + self.headerSize, stop + self.headerSize
            # Read in the bytes corresponding to the node to load and
            # decompress it
            dump = json.loads(lzma.decompress(_readBytes(start, stop)))

            # Turn the json back into the pydandtic model
            nodeDeserialized = SerializedQuantumNode.direct(**dump)
            del dump

            # attach the dictionary of dimension records to the pydantic model
            # these are stored separately because the are stored over and over
            # and this saves a lot of space and time.
            nodeDeserialized.quantum.dimensionRecords = self.infoMappings.dimensionRecords
            # get the label for the current task
            nodeTaskLabel = nodeDeserialized.taskLabel

            if nodeTaskLabel not in loadedTaskDef:
                # Get the byte ranges corresponding to this taskDef
                start, stop = self.infoMappings.taskDefMap[nodeTaskLabel]["bytes"]
                start, stop = start + self.headerSize, stop + self.headerSize

                # bytes are compressed, so decompress them
                taskDefDump = json.loads(lzma.decompress(_readBytes(start, stop)))
                taskClass: type[PipelineTask] = doImportType(taskDefDump["taskName"])
                config: PipelineTaskConfig = taskClass.ConfigClass()
                config.loadFromStream(taskDefDump["config"])
                # Rebuild TaskDef
                recreatedTaskDef = TaskDef(
                    taskName=taskDefDump["taskName"],
                    taskClass=taskClass,
                    config=config,
                    label=taskDefDump["label"],
                )
                loadedTaskDef[nodeTaskLabel] = recreatedTaskDef

                # initInputRefs and initOutputRefs are optional
                if (refs := taskDefDump.get("initInputRefs")) is not None:
                    initInputRefs[recreatedTaskDef] = [
                        cast(DatasetRef, DatasetRef.from_json(ref, universe=universe)) for ref in refs
                    ]
                if (refs := taskDefDump.get("initOutputRefs")) is not None:
                    initOutputRefs[recreatedTaskDef] = [
                        cast(DatasetRef, DatasetRef.from_json(ref, universe=universe)) for ref in refs
                    ]

                # rebuild the mappings that associate dataset type names with
                # TaskDefs
                for _, input in self.infoMappings.taskDefMap[nodeTaskLabel]["inputs"]:
                    datasetDict.addConsumer(DatasetTypeName(input), recreatedTaskDef)

                added = set()
                for outputConnection in self.infoMappings.taskDefMap[nodeTaskLabel]["outputs"]:
                    typeName = outputConnection[1]
                    if typeName not in added:
                        added.add(typeName)
                        datasetDict.addProducer(DatasetTypeName(typeName), recreatedTaskDef)

            # reconstitute the node, passing in the dictionaries for the
            # loaded TaskDefs and dimension records. These are used to ensure
            # that each unique record is only loaded once
            qnode = QuantumNode.from_simple(nodeDeserialized, loadedTaskDef, universe)
            container[qnode.nodeId] = qnode
            taskToQuantumNode[loadedTaskDef[nodeTaskLabel]].add(qnode)

            # recreate the relations between each node from stored info
            graph.add_node(qnode)
            for id in self.infoMappings.map[qnode.nodeId]["inputs"]:
                # uuid is stored as a string, turn it back into a uuid
                id = uuid.UUID(id)
                # if the id is not yet in the container, dont make a connection
                # this is not an issue, because once it is, that id will add
                # the reverse connection
                if id in container:
                    graph.add_edge(container[id], qnode)
            for id in self.infoMappings.map[qnode.nodeId]["outputs"]:
                # uuid is stored as a string, turn it back into a uuid
                id = uuid.UUID(id)
                # if the id is not yet in the container, dont make a connection
                # this is not an issue, because once it is, that id will add
                # the reverse connection
                if id in container:
                    graph.add_edge(qnode, container[id])

        newGraph = object.__new__(QuantumGraph)
        newGraph._metadata = self.infoMappings.metadata
        newGraph._buildId = self.infoMappings._buildId
        newGraph._datasetDict = datasetDict
        newGraph._nodeIdMap = container
        newGraph._count = len(nodes)
        newGraph._taskToQuantumNode = dict(taskToQuantumNode.items())
        newGraph._taskGraph = datasetDict.makeNetworkXGraph()
        newGraph._connectedQuanta = graph
        newGraph._initInputRefs = initInputRefs
        newGraph._initOutputRefs = initOutputRefs
        newGraph._globalInitOutputRefs = self.infoMappings.globalInitOutputRefs
        newGraph._registryDatasetTypes = self.infoMappings.registryDatasetTypes
        newGraph._universe = universe
        newGraph._pipeline_graph = None
        return newGraph


DESERIALIZER_MAP: dict[int, type[DeserializerBase]] = {
    1: DeserializerV1,
    2: DeserializerV2,
    3: DeserializerV3,
}



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining few methods to manipulate or query pipelines.
"""

from __future__ import annotations

# No one should do import * from this module
__all__ = ["isPipelineOrdered", "orderPipeline"]

import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING

from deprecated.sphinx import deprecated
from lsst.utils.introspection import find_outside_stacklevel

from .pipeline import Pipeline, TaskDef

# Exceptions re-exported here for backwards compatibility.
from .pipeline_graph import DuplicateOutputError, PipelineDataCycleError, PipelineGraph  # noqa: F401

if TYPE_CHECKING:
    from .taskFactory import TaskFactory

# TODO: remove this module on DM-40443.
warnings.warn(
    "The pipeTools module and its contents are deprecated in favor of PipelineGraph, and will be removed "
    "after v27.",
    category=FutureWarning,
    stacklevel=find_outside_stacklevel("lsst.pipe.base"),
)


@deprecated(
    "Deprecated and will be removed after v27.",
    version="v27.0",
    category=FutureWarning,
)
class MissingTaskFactoryError(Exception):
    """Exception raised when client fails to provide TaskFactory instance."""

    pass


@deprecated(
    "Deprecated in favor of PipelineGraph methods and will be removed after v27.",
    version="v27.0",
    category=FutureWarning,
)
def isPipelineOrdered(pipeline: Pipeline | Iterable[TaskDef], taskFactory: TaskFactory | None = None) -> bool:
    """Check whether tasks in pipeline are correctly ordered.

    Pipeline is correctly ordered if for any DatasetType produced by a task
    in a pipeline all its consumer tasks are located after producer.

    Parameters
    ----------
    pipeline : `Pipeline` or `collections.abc.Iterable` [ `TaskDef` ]
        Pipeline description.
    taskFactory : `TaskFactory`, optional
        Ignored; present only for backwards compatibility.

    Returns
    -------
    is_ordered : `bool`
        True for correctly ordered pipeline, False otherwise.

    Raises
    ------
    ImportError
        Raised when task class cannot be imported.
    DuplicateOutputError
        Raised when there is more than one producer for a dataset type.
    """
    if isinstance(pipeline, Pipeline):
        graph = pipeline.to_graph()
    else:
        graph = PipelineGraph()
        for task_def in pipeline:
            graph.add_task(task_def.label, task_def.taskClass, task_def.config, task_def.connections)
    # Can't use graph.is_sorted because that requires sorted dataset type names
    # as well as sorted tasks.
    tasks_xgraph = graph.make_task_xgraph()
    seen: set[str] = set()
    for task_label in tasks_xgraph:
        successors = set(tasks_xgraph.successors(task_label))
        if not successors.isdisjoint(seen):
            return False
        seen.add(task_label)
    return True


@deprecated(
    "Deprecated in favor of PipelineGraph methods and will be removed after v27.",
    version="v27.0",
    category=FutureWarning,
)
def orderPipeline(pipeline: Pipeline | Iterable[TaskDef]) -> list[TaskDef]:
    """Re-order tasks in pipeline to satisfy data dependencies.

    Parameters
    ----------
    pipeline : `Pipeline` or `collections.abc.Iterable` [ `TaskDef` ]
        Pipeline description.

    Returns
    -------
    ordered : `list` [ `TaskDef` ]
        Correctly ordered pipeline.

    Raises
    ------
    DuplicateOutputError
        Raised when there is more than one producer for a dataset type.
    PipelineDataCycleError
        Raised when the pipeline has dependency cycles.
    """
    if isinstance(pipeline, Pipeline):
        graph = pipeline.to_graph()
    else:
        graph = PipelineGraph()
        for task_def in pipeline:
            graph.add_task(task_def.label, task_def.taskClass, task_def.config, task_def.connections)
    graph.sort()
    return list(graph._iter_task_defs())
